#!/usr/bin/python3
# type: ignore

# FIXME:
# make tag transfer into something like the encoder registry
# but simpler -- either direct transfers with shortcut functions,
# or loaders that load generic info, and writers of that generic
# info for known formats.
#
# IOW: this must be rewritten completely.

from __future__ import print_function

import logging
import os
import typing

import mutagen.apev2  # atype: ignore
import mutagen.flac  # atype: ignore
import mutagen.id3  # atype: ignore
import mutagen.musepack  # atype: ignore
import mutagen.oggvorbis  # atype: ignore

from .util import transform_keys

logger = logging.getLogger(__name__)


def copy_generic_tag_values_to_id3(i, o):
    """copies tag values from i to o, transforming to fit the ID3 spec"""
    tag_transformation_map = {
        "album": "TALB",
        "artist": "TPE1",
        "comment": "COMM",
        "description": "COMM",
        "date": "TDRC",
        "year": "TDRC",
        "title": "TIT2",
        "tracknumber": "TRCK",
        "track": "TRCK",
        "genre": "TCON",
        "replaygain_album_gain": lambda x: mutagen.id3.TXXX(
            desc="replaygain_album_gain", encoding=1, text=x
        ),
        "replaygain_album_peak": lambda x: mutagen.id3.TXXX(
            desc="replaygain_album_peak", encoding=1, text=x
        ),
        "replaygain_track_gain": lambda x: mutagen.id3.TXXX(
            desc="replaygain_track_gain", encoding=1, text=x
        ),
        "replaygain_track_peak": lambda x: mutagen.id3.TXXX(
            desc="replaygain_track_peak", encoding=1, text=x
        ),
    }

    def iterate(k, v):
        if k in tag_transformation_map:
            if type(v) not in (tuple, list):
                v = [v]
            constructor = tag_transformation_map[k]
            if callable(constructor):
                newvalues = [constructor(str(value)) for value in v]
            else:
                constructor = getattr(mutagen.id3, constructor)
                newvalues = [constructor(text=str(value), encoding=3) for value in v]
            list(map(o.add, newvalues))
            return newvalues

    transform_keys(str.lower, i)
    for k, v in list(i.items()):
        iterate(k, v)


def transfer_tags_any_mp3(origin, destination, tag_reader):

    try:
        i = tag_reader(origin)
    except Exception:
        logger.exception("Could not open tag of %s", origin)
        return

    try:
        o = mutagen.id3.ID3(filename=destination)
    except mutagen.id3.ID3NoHeaderError:
        o = mutagen.id3.ID3()

    copy_generic_tag_values_to_id3(i, o)
    o.save(filename=destination, v1=2)


def transfer_tags_mp3(origin, destination):

    # First we'll try to copy any APEv2 tags from origin to destination.
    # It is not an error if that does not happen.
    try:
        i = mutagen.apev2.APEv2()
        i.load(origin)
        i.save(destination)
    except mutagen.apev2.APENoHeaderError:
        pass
    except Exception:
        logger.exception(
            "Though %s exists, an error blocked opening the APEv2 tag", origin
        )

    try:
        i = mutagen.id3.ID3()
        i.load(origin)
        i.save(destination, v1=2)
    except mutagen.id3.ID3NoHeaderError:
        pass  # no ID3 tag, we skip it
    except Exception:
        logger.exception(
            "Though %s exists, an error blocked opening the ID3 tag: %s", origin
        )


tag_transfer_functions = {
    "mp3:mp3": lambda x, y, _, __: transfer_tags_mp3(x, y),
    "mp3:aac": lambda x, y, _, __: transfer_tags_mp3(x, y),
    "ogg:mp3": lambda x, y, _, __: transfer_tags_any_mp3(x, y, mutagen.oggvorbis.Open),
    "flac:mp3": lambda x, y, _, __: transfer_tags_any_mp3(x, y, mutagen.flac.Open),
    "mpc:mp3": lambda x, y, _, __: transfer_tags_any_mp3(x, y, mutagen.musepack.Open),
}


def transfer_tags(
    origin: str,
    destination: str,
    source_format: typing.Optional[str] = None,
    target_format: typing.Optional[str] = None,
) -> None:
    if not source_format:
        source_format = os.path.splitext(origin)[1][1:].lower()
    if not target_format:
        target_format = os.path.splitext(destination)[1][1:].lower()
    m = "%s:%s" % (source_format, target_format)
    if m in tag_transfer_functions:
        logger.debug("Using %s for tag transfer function", m)
        transfer_function = tag_transfer_functions[m]
        transfer_function(origin, destination, source_format, target_format)
    else:
        logger.debug("No tag transfer function for %s", m)
    return



import logging
import typing


class CLIFormatter(logging.Formatter):
    def __init__(
        self,
        main_module_name: str,
        fmt: typing.Optional[str] = "%(levelname)s: (%(name)s) \t%(message)s",
    ) -> None:
        self.main_module_name = main_module_name
        self.normal_formatter = logging.Formatter(fmt)
        self.main_formatter = logging.Formatter("%(message)s")

    def format(self, record: logging.LogRecord) -> str:
        f = (
            self.main_formatter
            if (record.name == self.main_module_name and record.levelno == logging.INFO)
            else self.normal_formatter
        )
        return f.format(record)


def basicConfig(main_module_name: str, level: int) -> None:
    handler = logging.StreamHandler()
    handler.setFormatter(CLIFormatter(main_module_name))
    logging.basicConfig(handlers=[handler], level=level)



from pathlib import Path
import shutil
from typing import List

from . import base
from .. import registry
from ..interfaces import FileType


@registry.register
class Copy(base.NoSettings):
    """Copies files without changing format."""

    cost = 1

    def transcode(self, src: Path, dst: Path) -> None:
        shutil.copyfile(src, dst)

    def can_transcode(self, src: Path) -> List[FileType]:
        return [FileType.from_path(src)]



from pathlib import Path
import unittest

from musictoolbox.transcoding.interfaces import FileType, TranscoderName
from musictoolbox.transcoding.policies import select_pipelines, TranscoderPolicy
from musictoolbox.transcoding.test_registry import mp


bn = FileType.by_name


class TestSelectPipeline(unittest.TestCase):
    def setUp(self) -> None:
        self.pipelines = [
            mp(1, [("mp3", "mp3", "copy")]),
            mp(2, [("mp3", "ogg", "mp3toogg")]),
            mp(3, [("mp3", "wav", "mp3towav"), ("wav", "ogg", "wavtoogg")]),
        ]

    def test_copy_selected(self) -> None:
        res = select_pipelines(
            self.pipelines,
            Path("a.mp3"),
            pipeline=[TranscoderName("copy")],
            dsttypes=[],
        )
        if not res:
            assert 0, "expected at least one pipeline"
        p = res[0]
        self.assertEqual([x.transcoder_name for x in p.steps], [TranscoderName("copy")])

    def test_cheaper_selected(self) -> None:
        res = select_pipelines(
            self.pipelines, Path("a.mp3"), dsttypes=[FileType.by_name("ogg")]
        )
        if not res:
            assert 0, "expected at least one pipeline"
        p = res[0]
        self.assertEqual(
            [x.transcoder_name for x in p.steps], [TranscoderName("mp3toogg")]
        )

    def test_pipeline_preference_respected(self) -> None:
        res = select_pipelines(
            self.pipelines,
            Path("a.mp3"),
            dsttypes=[FileType.by_name("ogg")],
            pipeline=[TranscoderName("mp3towav"), TranscoderName("wavtoogg")],
        )
        if not res:
            assert 0, "expected at least one pipeline"
        p = res[0]
        self.assertEqual(
            [x.transcoder_name for x in p.steps],
            [TranscoderName("mp3towav"), TranscoderName("wavtoogg")],
        )

    def test_pipeline_with_transcode_to(self) -> None:
        res = select_pipelines(
            self.pipelines,
            Path("a.mp3"),
            dsttypes=[FileType.by_name("wav"), FileType.by_name("ogg")],
            pipeline=[TranscoderName("mp3towav"), TranscoderName("wavtoogg")],
        )
        if not res:
            assert 0, "expected at least one pipeline"
        p = res[0]
        self.assertEqual(
            [x.transcoder_name for x in p.steps],
            [TranscoderName("mp3towav"), TranscoderName("wavtoogg")],
        )

    def test_no_pipeline_if_none_fit(self) -> None:
        res = select_pipelines(
            self.pipelines,
            Path("a.mp3"),
            dsttypes=[FileType.by_name("mp4")],
        )
        if res:
            assert 0, "expected no pipelines, got %s" % res


class TestTranscoderPolicy(unittest.TestCase):
    def test_one(self) -> None:
        p = TranscoderPolicy(bn("mp3"), bn("mov"), None, None)
        assert p.match(bn("mp3"))
        assert p.match(dsttype=bn("mov"))
        assert p.match(srctype=bn("mp3"), dsttype=bn("mov"))

    def test_only_target_policy(self) -> None:
        p = TranscoderPolicy(None, bn("mov"), None, None)
        assert p.match(bn("mp3"))
        assert p.match(dsttype=bn("mov"))
        assert not p.match(dsttype=bn("mp4"))
        assert p.match(srctype=bn("mp3"), dsttype=bn("mov"))



import argparse
import collections
import itertools
import logging
import subprocess
import typing
import concurrent.futures
import os

from musictoolbox.cache import FileMetadataCache, OnDiskMetadataCache
from musictoolbox.files import all_files
from musictoolbox.logging import basicConfig
from mutagen import File
from rgain3.albumid import get_album_id
from rgain3 import rgio, GainData


_LOGGER = logging.getLogger(__name__)
_MAPPER = rgio.BaseFormatsMap()

CACHE_VERSION = 6

TM = typing.TypeVar("TM", bound="AlbumIdentifier")


class AlbumIdentifier(object):
    identifier: str
    albumgain: GainData | None
    trackgain: GainData | None

    def __init__(
        self,
        valid: bool,
        identifier: str,
        albumgain: GainData | None,
        trackgain: GainData | None,
    ) -> None:
        self.valid = valid
        self.identifier = identifier
        self.albumgain = albumgain
        self.trackgain = trackgain

    @classmethod
    def from_file(klass: typing.Type[TM], ff: str) -> TM:
        from_disk_metadata = None
        try:
            from_disk_metadata = File(ff)
        except Exception as exc:
            _LOGGER.error("Error identifying %s: %s:", ff, exc)
            return klass(False, "", None, None)
        if from_disk_metadata is None:
            return klass(False, "", None, None)
        album_id = get_album_id(from_disk_metadata)
        try:
            trackgain, albumgain = _MAPPER.read_gain(ff)
        except Exception as exc:
            # The file is not supported.  We return invalid.
            _LOGGER.error("Cannot read ReplayGain from %s: %s>", ff, exc)
            return klass(False, "", None, None)
        return klass(
            True,
            identifier=album_id,
            trackgain=trackgain,
            albumgain=albumgain,
        )


def main() -> int:
    p = argparse.ArgumentParser(
        description="Collection-wide frontend for replaygain (of python-rgain3 fame).",
        epilog="This program"
        " helps you add ReplayGain information to music in an intelligent fashion,"
        " grouping songs by album so album ReplayGain information is calculated"
        " correctly with respect to all other tracks in the album."
        "\n\n"
        "This program is similar to the rgain3 program `collectiongain` but it"
        " decides which files to group as an album in a different way, and will not"
        " operate recursively by default."
        "\n\n"
        "To ensure optimum operation, use the program scanalbumartists prior to"
        " running this program, to ensure all your album artist tags are consistent.",
    )
    p.add_argument("FILE", type=str, nargs="+", help="file to operate on")
    p.add_argument(
        "--show",
        action="store_true",
        help="only show the ReplayGain information for files",
    )
    p.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        help="if folders are specified, operate recursively",
    )
    p.add_argument(
        "-n",
        "--dry-run",
        action="store_true",
        help="do not actually modify files",
    )
    p.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="force recalculation even if files already contain that information",
    )
    p.add_argument(
        "-p",
        "--parallelism",
        type=int,
        help="how many replaygain processes to run at once (default %(default)s)",
        default=os.cpu_count(),
    )
    p.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="enable verbose operation",
    )

    args = p.parse_args()
    basicConfig(
        main_module_name=__name__, level=logging.DEBUG if args.verbose else logging.INFO
    )

    allfiles = all_files(args.FILE, recursive=args.recursive)

    with OnDiskMetadataCache(
        "doreplaygain.pickle",
        CACHE_VERSION,
        lambda: FileMetadataCache(AlbumIdentifier.from_file),
    ) as tags:
        tags.update_metadata_for(allfiles)

        album_files: dict[str, list[str]] = collections.defaultdict(list)

        for f in allfiles:
            id_ = tags.get(f)
            if not id_ or not id_.valid:
                continue
            album_files[id_.identifier].append(f)

        ret = 0

        if args.show:
            if not album_files:
                return ret
            for album, files in album_files.items():
                if album is None:
                    _LOGGER.info(f"* For singles {album}:")
                else:
                    _LOGGER.info(f"* For album {album}:")
                r = subprocess.call(["replaygain", "--show"] + files)
                if r != 0:
                    ret = r
            return ret

        t = concurrent.futures.ThreadPoolExecutor(max_workers=args.parallelism)
        future_to_result = {}

        for album, filegroup in album_files.items():
            if not album:
                batches = [[x] for x in filegroup]
            else:
                batches = [filegroup]
            for batch in batches:
                process = False
                noalbum = not album or len(batch) < 2
                if noalbum:
                    # If the file in the batch and has RG, ignore the batch
                    # unless explicitly instructed not to.
                    if args.force or any(tags[x].trackgain is None for x in batch):
                        process = True
                else:
                    # If all of the files file in the batch have equal album RG
                    # ignore the batch unless explicitly instructed not to.
                    if args.force or any(
                        z != w
                        for z, w in itertools.pairwise(tags[x].albumgain for x in batch)
                    ):
                        process = True
                if not process:
                    continue
                future_to_result[
                    t.submit(
                        subprocess.call,
                        ["replaygain"]
                        + (["--dry-run"] if args.dry_run else [])
                        + ["--force"]
                        + (["--no-album"] if noalbum else [])
                        + batch,
                    )
                ] = batch

        try:
            for future in concurrent.futures.as_completed(future_to_result):
                batch = future_to_result[future]
                ret = future.result()
                if ret == 0:
                    tags.update_metadata_for(batch)
                else:
                    _LOGGER.error("Batch %s failed", batch)
                    for future in future_to_result:
                        future.cancel()
                    break
        except BaseException:
            for future in future_to_result:
                future.cancel()
            raise

        return ret



import argparse
import collections
import itertools
import logging
import collections.abc
import os
import sys
import typing
from musictoolbox.logging import basicConfig
from musictoolbox.cache import FileMetadataCache, OnDiskMetadataCache
from musictoolbox.files import all_files

from mutagen import File

_LOGGER = logging.getLogger(__name__)

KEY_ALBUM = "album"
KEY_ALBUMARTIST = "albumartist"
KEY_ARTIST = "artist"
CACHE_VERSION = 3

TM = typing.TypeVar("TM", bound="TrackMetadata")


class TrackMetadata(object):
    album: list[str] | None
    artist: list[str] | None
    albumartist: list[str] | None

    def __init__(
        self,
        valid: bool,
        album: list[str],
        artist: list[str],
        albumartist: list[str],
    ) -> None:
        self.valid = valid
        self.album = album
        self.albumartist = albumartist
        self.artist = artist

    @classmethod
    def from_file(klass: typing.Type[TM], ff: str) -> TM:
        try:
            from_disk_metadata = File(ff, easy=True)
        except Exception as exc:
            _LOGGER.error("Error identifying %s: %s:", ff, exc)
            from_disk_metadata = None
        if from_disk_metadata is None:
            return klass(False, [], [], [])
        return klass(
            True,
            album=from_disk_metadata.get(KEY_ALBUM),
            artist=from_disk_metadata.get(KEY_ARTIST),
            albumartist=from_disk_metadata.get(KEY_ALBUMARTIST),
        )


def main() -> int:
    p = argparse.ArgumentParser(
        description="Determine which albums exist and if they are properly tagged."
    )
    p.add_argument("FILE", type=str, nargs="+", help="file to operate on")
    p.add_argument(
        "-f",
        "--fix",
        action="store_true",
        help="add a various artists album artist tag to files that need it",
    )
    p.add_argument(
        "-t",
        "--various-artists-tag",
        help="content of the various artists album artist tag;"
        " if this is set to auto, then the album artist is deduced"
        " from the most popular artist among the album",
        default="Various artists",
    )
    p.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="enable verbose operation",
    )
    args = p.parse_args()

    basicConfig(
        main_module_name=__name__,
        level=(logging.DEBUG if args.verbose else logging.INFO),
    )

    allfiles = all_files(args.FILE, recursive=True)

    with OnDiskMetadataCache(
        "scanalbumartists.pickle",
        CACHE_VERSION,
        lambda: FileMetadataCache(TrackMetadata.from_file),
    ) as tags:
        tags.update_metadata_for(allfiles)

        album_tree: dict[
            str, dict[str | None, dict[str | None, dict[str | None, list[str]]]]
        ] = collections.defaultdict(
            lambda: collections.defaultdict(
                lambda: collections.defaultdict(lambda: collections.defaultdict(list))
            )
        )

        for fn in allfiles:
            tag = tags.get(fn)
            if not tag:
                continue

            folder = os.path.dirname(fn)
            album = tag.album[0] if tag.album else None
            if not album:
                continue
            artist = tag.artist[0] if tag.artist else None
            albumartist = tag.albumartist[0] if tag.albumartist else None
            try:
                album_tree[folder][album][albumartist][artist].append(fn)
            except Exception:
                assert 0, (folder, album, albumartist, artist, fn)

        for folder, data in sorted(album_tree.items()):
            for album, albumartists_data in data.items():
                problems = []
                if None in albumartists_data and len(albumartists_data[None]) > 1:
                    problems.append(
                        "This album has more than one artist but no album artist"
                    )
                if len(albumartists_data) > 1:
                    problems.append("This album has more than one album artist")
                if problems:
                    print(f"{folder}:")
                    print(f"  {album}:")
                    for problem in problems:
                        print(f"    {problem}")
                    files_to_fix = [
                        f
                        for _, artistdata in albumartists_data.items()
                        for _, artist_files in artistdata.items()
                        for f in artist_files
                    ]

                    def dumpfiles(header: str) -> None:
                        print(f"    {header}:")
                        for f in files_to_fix:
                            print(f"      {f}")

                    va_text = args.various_artists_tag
                    if va_text == "auto":
                        artists = [
                            "" if a is None else a
                            for _, artistdata in albumartists_data.items()
                            for a, artist_files in artistdata.items()
                            for _ in artist_files
                        ]
                        popular_artists = sorted(
                            [
                                (x, len(list(y)))
                                for x, y in itertools.groupby(sorted(artists))
                            ],
                            key=lambda g: -g[1],
                        )
                        va_text = popular_artists[0][0]
                        if (
                            len(popular_artists) > 1
                            and popular_artists[0][1] == popular_artists[1][1]
                        ):
                            va_text = None
                    if va_text is None:
                        dumpfiles("Cannot determine album artist, skipping these files")
                        continue

                    if args.fix:
                        print(
                            f"    Adding album artist {va_text} among {popular_artists}"
                        )
                        for f in files_to_fix:
                            savetag = File(f, easy=True)
                            savetag[KEY_ALBUMARTIST] = [va_text]
                            savetag.save()
                        tags.update_metadata_for(files_to_fix)
                        dumpfiles("Files fixed")
                    else:
                        dumpfiles("Files to fix")
                        print(
                            f"    Would add the album artist {va_text}"
                            f" among {popular_artists}"
                        )

    return 0


if __name__ == "__main__":
    sys.exit(main())



import typing

AnyIn = typing.TypeVar("AnyIn")
AnyOut = typing.TypeVar("AnyOut")


def transform_keys(
    f: typing.Callable[[AnyIn], AnyIn], d: typing.Dict[AnyIn, AnyOut]
) -> typing.Dict[AnyIn, AnyOut]:
    for k, v in list(d.items()):
        newk = f(k)
        if newk != k:
            del d[k]
            d[newk] = v
    return d


__all__ = ["transform_keys"]



import glob
import logging
import os
from pathlib import Path
import re
import subprocess
from typing import Union, Optional, Any, Callable, List, Tuple, Dict

import packaging.version

from . import base
from .. import registry
from ..interfaces import FileType


try:
    from shlex import quote
except ImportError:
    from pipes import quote

try:
    from urllib.request import pathname2url
except ImportError:
    from urllib import pathname2url  # type:ignore

logger = logging.getLogger(__name__)


_gst_command = None


def sort_gst_candidates(candidates: List[str]) -> List[str]:
    """Sort GStreamer launch candidates, highest versions first."""
    version_re = re.compile(".*-([0-9]+($|[.][0-9]+)+)")

    def getver(
        v: str,
    ) -> Union[None, packaging.version.Version, packaging.version.LegacyVersion]:
        try:
            return packaging.version.parse(v)
        except packaging.version.InvalidVersion:
            return None

    withversions = [
        (
            version_re.match(name).groups()[0]  # type:ignore
            if (version_re.match(name) is not None and version_re.match(name))
            else "10000000",
            name,
        )
        for name in candidates
    ]
    withversions = [(getver(v), x) for v, x in withversions if getver(v)]
    return [x[1] for x in reversed(sorted(withversions))]


def gst(
    src: Path,
    dst: Path,
    *elements: Union[str, List[str]],
    force_gst_command: Optional[str] = None
) -> List[str]:
    if force_gst_command:
        prog = force_gst_command
    else:
        global _gst_command
        if _gst_command is None:
            prog = "gst-launch-1.0"
            for p in os.environ.get("PATH", "").split(os.path.pathsep):
                opts = sort_gst_candidates(
                    glob.glob(os.path.join(glob.escape(p), "gst-launch*"))
                )
                for opt in opts:
                    if os.access(opt, os.X_OK):
                        prog = opt
                        break
                else:
                    continue
                break
            _gst_command = prog
        prog = _gst_command

    cmd = [prog, "-f"]
    srcs = "file://" + pathname2url(src.absolute().as_posix())
    cmd += ["giosrc", "location=%s" % srcs]
    for element in elements:
        cmd.append("!")
        if isinstance(element, str):
            cmd.append(element)
        elif isinstance(element, list):
            for e in element:
                if isinstance(e, str):
                    cmd.append(e)
                else:
                    assert 0
        else:
            assert 0
    cmd.append("!")
    cmd += [
        "filesink",
        "location=%s" % dst.absolute().as_posix(),
    ]
    return cmd


def get_output(cmd: List[str]) -> str:
    logger.debug("Getting output from %s", " ".join(quote(s) for s in cmd))
    return subprocess.check_output(cmd, stderr=subprocess.STDOUT, text=True)


def run(cmd: List[str], env: Dict[str, str] = {}) -> None:
    logger.debug("Running %s", " ".join(quote(s) for s in cmd))
    newenv = dict(os.environ.items())
    for k, v in env.items():
        newenv[k] = v
    output = subprocess.check_output(
        cmd,
        stdin=None,
        stderr=subprocess.STDOUT,
        close_fds=True,
        env=newenv,
    )
    if "gst-launch" in cmd[0] and b"Got EOS from element" not in output:
        # Artificially raise an error here, as there was a problem with
        # the pipeline not finishing.  Notorious for being necessary
        # when gst-launch is interrupted, and it still returns zero.
        raise subprocess.CalledProcessError(255, cmd, "", output)


@registry.register
class FlvMp4WebmToMp3(base.WithEnvironmentVariables):
    """Transcodes from FLV / MP4 to MP3, avoiding retranscoding if possible."""

    cost = 20

    def can_transcode(self, src: Path) -> List[FileType]:
        return (
            [FileType.by_name("mp3")]
            if FileType.from_path(src) in ["flv", "mp4", "webm", "mkv"]
            else []
        )

    def transcode(self, src: Path, dst: Path) -> None:
        """Transcode FLV / MP4 to MP3 file"""
        output = get_output(["ffprobe", src.as_posix()])
        if "Audio: mp3" in output:
            cmd = gst(src, dst, "flvdemux", "audio/mpeg")
        else:
            cmd = gst(
                src,
                dst,
                "decodebin",
                "audioconvert",
                [
                    "lamemp3enc",
                    "encoding-engine-quality=2",
                    "quality=0",
                ],
            )
        cmd.append("xingmux")
        run(cmd, env=self.settings["environment_variables"])


@registry.register
class ExtractAudio(object):
    """Extracts audio from videos into a suitably-encapsulated file."""

    cost = 3

    def __init__(self, settings: Optional[Dict[str, Any]]):
        self.settings = settings or {}
        defaults: List[Tuple[str, Callable[[Any], bool], Any]] = [
            (
                "source_extensions",
                lambda v: isinstance(v, list),
                ["flv", "mp4", "webm", "mkv", "wmv", "m4v"],
            ),
        ]
        for k, t, v in defaults:
            if k not in self.settings:
                self.settings[k] = self.settings.get(k, v)
                if not t(self.settings[k]):
                    raise ValueError("Invalid setting %s: %s" % (k, v))

    def can_transcode(self, src: Path) -> List[FileType]:
        if not src.is_file():
            # Optimization.  Cannot inspect for transcoding what does not exist.
            return []

        if (
            FileType.from_path(src) in self.settings["source_extensions"]
        ) or "*" in self.settings["source_extensions"]:
            pass
        else:
            # Skip this file.  As configured, it is not supported by this encoder.
            return []

        try:
            output = get_output(
                [
                    "ffprobe",
                    "-loglevel",
                    "warning",
                    "-hide_banner",
                    "-show_streams",
                    "-select_streams",
                    "a",
                    "--",
                    src.as_posix(),
                ]
            )
        except subprocess.CalledProcessError:
            # Cannot detect file.
            return []
        types: List[FileType] = []
        if "codec_name=mp3" in output:
            types.append(FileType.by_name("mp3"))
        if "codec_name=aac" in output:
            types.append(FileType.by_name("m4a"))
        if "codec_name=opus" in output:
            types.append(FileType.by_name("opus"))
        if "codec_name=vorbis" in output:
            types.append(FileType.by_name("ogg"))
        if "codec_name=wmav" in output:
            types.append(FileType.by_name("wma"))
        return types

    def transcode(self, src: Path, dst: Path) -> None:
        cmd = [
            "ffmpeg",
            "-loglevel",
            "quiet",
            "-y",
            "-i",
            src.as_posix(),
            "-acodec",
            "copy",
            "-vn",
            dst.as_posix(),
        ]
        run(cmd)


@registry.register
class FlvMp4WebmToWav(base.WithEnvironmentVariables):
    """Transcodes from FLV / MP4 to RIFF WAVE 32 bit float."""

    cost = 10

    def can_transcode(self, src: Path) -> List[FileType]:
        return (
            [FileType.by_name("wav")]
            if FileType.from_path(src) in ["flv", "mp4", "webm", "mkv"]
            else []
        )

    def transcode(self, src: Path, dst: Path) -> None:
        cmd = gst(
            src,
            dst,
            "decodebin",
            "audioconvert",
            "audio/x-raw,format=F32LE",
            "wavenc",
        )
        run(cmd, env=self.settings["environment_variables"])


@registry.register
class AudioToMp3(base.WithEnvironmentVariables):
    """Transcodes from any audio format to MP3."""

    cost = 10

    def can_transcode(self, src: Path) -> List[FileType]:
        return (
            [FileType.by_name("mp3")]
            if FileType.from_path(src)
            in ["ogg", "aac", "m4a", "wav", "flac", "mpc", "wma"]
            else []
        )

    def transcode(self, src: Path, dst: Path) -> None:
        cmd = gst(
            src,
            dst,
            "decodebin",
            "audioconvert",
            [
                "lamemp3enc",
                "encoding-engine-quality=2",
                "quality=0",
            ],
            "xingmux",
        )
        run(cmd, env=self.settings["environment_variables"])


@registry.register
class AudioToWav(base.NoSettings):
    """Transcodes from any audio format to RIFF WAVE 32 bit float."""

    cost = 10

    def can_transcode(self, src: Path) -> List[FileType]:
        return (
            [FileType.by_name("wav")]
            if FileType.from_path(src)
            in ["ogg", "aac", "m4a", "flac", "mpc", "mp3", "wma", "opus"]
            else []
        )

    def transcode(self, src: Path, dst: Path) -> None:
        cmd = gst(
            src,
            dst,
            "decodebin",
            "audioconvert",
            "audio/x-raw,format=F32LE",
            "wavenc",
        )
        run(cmd)


@registry.register
class WavToOgg(base.NoSettings, base.BinaryTranscoder):
    """Transcodes from any audio format to Ogg Vorbis."""

    cost = 10

    def transcode(self, src: Path, dst: Path) -> None:
        cmd = gst(
            src,
            dst,
            "wavparse",
            "audioconvert",
            [
                "vorbisenc",
                "quality=0.49",
            ],
            "oggmux",
        )
        run(cmd)


@registry.register
class WavToOpus(base.NoSettings, base.BinaryTranscoder):
    """Transcodes from any audio format to Ogg Opus."""

    cost = 9

    def transcode(self, src: Path, dst: Path) -> None:
        cmd = gst(
            src,
            dst,
            "wavparse",
            "audioconvert",
            ["audioresample", "quality=10", "sinc-filter-mode=full"],
            "opusenc",
            "oggmux",
        )
        run(cmd)



# FIXME: the following import should be done with a zope interfaces registry,
# and the Zope component architecture, since  the classes have already been
# registered as ITranscoder.

from . import basic, gstreamerffmpeg  # @UnusedImport



import argparse
import collections
import errno
import fnmatch
import os
import stat
import sys
import tempfile
import textwrap
import urllib.parse


def get_parser() -> argparse.ArgumentParser:
    program_name = os.path.basename(sys.argv[0])
    program_shortdesc = "Generates file lists / M3U playlists from a directory."
    program_shortdesc += (
        "\n\n"
        + textwrap.dedent(
            """
This program takes a number of paths in the command line, and a destination
file name, then saves the list of files within those paths to the destination
file, overwriting whatever was in the file previously.  Each path is only added
once to the destination file list / playlist.

The list of files in the file list / playlist created by `genplaylist` will
always be relative to the directory that contains the file list / playlist,
except for the case when a file has non-printable characters, in which case
a proper URL with the non-printable characters escaped will be saved to the
file list / playlist instead.

Only path entries that test positive for the `os.path.isfile()` test
(that means no special devices, no broken symlinks, no directories) will be
included in the output file list.  Unreadable files will be added, but
unlistable directories will be skipped without displaying any errors.
    """
        ).strip()
    )
    parser = argparse.ArgumentParser(
        prog=program_name,
        description=program_shortdesc,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-i",
        "--include",
        help="include only files matching this pattern; "
        "this option can be specified multiple times",
        metavar="pattern",
        action="append",
    )
    parser.add_argument(
        "-x",
        "--exclude",
        help="exclude files matching this pattern, as well as "
        "files within directories matching this pattern; "
        "this option can be specified multiple times",
        metavar="pattern",
        action="append",
    )
    parser.add_argument(
        dest="paths",
        help="paths to directories or files to list",
        metavar="path",
        nargs="+",
    )
    parser.add_argument(
        dest="playlist", help="destination playlist name", metavar="playlist"
    )
    return parser


def get_umask() -> int:
    umask = os.umask(0o77)
    os.umask(umask)
    return umask


class OutputWriter(object):
    """Writes to a regular file or a special file, varying its behavior
    depending on the type of file."""

    def __init__(self, filename: str) -> None:
        self.name = filename

        try:
            mode = os.stat(filename).st_mode
            if stat.S_ISBLK(mode) or stat.S_ISCHR(mode) or stat.S_ISFIFO(mode):
                self.fobject = open(filename, "w")
        except OSError:
            pass

        if not self.fobject:
            directory, prefix = os.path.split(filename)
            if not directory:
                directory = os.path.curdir
            if not prefix:
                raise OSError(
                    errno.EISDIR,
                    os.strerror(errno.EISDIR) + ": " + repr(filename),
                )
            t = tempfile.NamedTemporaryFile(
                mode="w+",
                prefix=prefix,
                dir=directory,
                delete=False,
            )
            t.close()
            self.fobject = open(t.name, "w")

    def write(self, text: str) -> None:
        self.fobject.write(text)

    def flush(self) -> None:
        self.fobject.flush()

    def close(self) -> None:
        try:
            if self.name != self.fobject.name:
                try:
                    oldmode = os.stat(self.name).st_mode
                    os.chmod(self.fobject.name, oldmode)
                except OSError as e:
                    if e.errno != errno.ENOENT:
                        raise
                    umask = get_umask()
                    newmode = ~umask & 0o666
                    os.chmod(self.fobject.name, newmode)
                os.rename(self.fobject.name, self.name)
        finally:
            try:
                self.fobject.close()
            except OSError as e:
                if e.errno != errno.ENOENT:
                    raise


def main() -> int:
    parser = get_parser()
    args = parser.parse_args()
    playlist = args.playlist
    excludes = args.exclude
    includes = args.include
    playlistdir = os.path.dirname(playlist)
    playlistfile = OutputWriter(playlist)
    playlistcontents = collections.OrderedDict()
    for path in args.paths:
        for basepath, dirs, files in os.walk(path, topdown=True):
            for d in dirs[:]:
                # Do not walk any directories that match exclusion patterns.
                if excludes and any(fnmatch.fnmatch(d, pat) for pat in excludes):
                    dirs.remove(d)
            for filename in files:
                # If includes were specified, do not include files that do
                # not match the inclusion patterns.
                if includes and not any(
                    fnmatch.fnmatch(filename, pat) for pat in includes
                ):
                    continue
                # Do not include any files that match exclusion patterns.
                if excludes and any(fnmatch.fnmatch(filename, pat) for pat in excludes):
                    continue
                joined = os.path.join(basepath, filename)
                if not os.path.isfile(joined):
                    continue
                if any(ord(x) < 32 for x in joined):
                    joined = os.path.abspath(joined)
                    p = "file://" + "".join(
                        x if x >= 32 and x < 128 else urllib.parse.quote(x)
                        for x in joined
                    )
                else:
                    p = os.path.relpath(joined, playlistdir)
                if p in playlistcontents:
                    continue
                playlistfile.write(p + "\n")
                playlistcontents[p] = True
    playlistfile.flush()
    playlistfile.close()

    return 0



import logging
import os
from pathlib import Path
import shutil
import tempfile
from threading import Lock
import typing

from networkx import MultiDiGraph  # type: ignore

from . import policies as pol, registry as reg
from .. import files
from .interfaces import (
    Postprocessor,
    FileType,
    TranscoderName,
)


logger = logging.getLogger(__name__)


class NoPipeline(Exception):
    def __init__(self, source: Path):
        self.source = source

    def __str__(self) -> str:
        return (
            "<Transcoding: file %s cannot be transcoded by any codec pipeline according to configuration>"
            % (self.source,)
        )


class TranscodingMapper(object):
    def __init__(
        self,
        transcoder_registry: reg.TranscoderRegistry,
        pipeline_selector: pol.PolicyBasedPipelineSelector,
    ):
        self.transcoder_registry = transcoder_registry
        self.pipeline_selector = pipeline_selector
        self.pipeline_cache: typing.Dict[Path, typing.List[reg.TranscodingPath]] = {}
        self.pipeline_cache_lock = Lock()

    def _feed_cache(self, path: Path) -> typing.List[reg.TranscodingPath]:
        with self.pipeline_cache_lock:
            if path in self.pipeline_cache:
                return self.pipeline_cache[path]

        _, all_paths = self.transcoder_registry.map_pipelines(path)
        transcoding_paths = self.pipeline_selector.select_pipelines(all_paths, path)
        with self.pipeline_cache_lock:
            self.pipeline_cache[path] = transcoding_paths
        return transcoding_paths

    def map(self, path: Path) -> Path:
        transcoding_paths = self._feed_cache(path)
        if not transcoding_paths:
            raise NoPipeline(path)
        transcoding_path = transcoding_paths[0]
        parent, stem = path.parent, path.stem
        newext = transcoding_path.dsttype
        dst = parent / (stem + "." + newext)
        return dst

    def lookup(self, path: Path) -> typing.List[reg.TranscodingPath]:
        return self._feed_cache(path)

    def lookup_with_graph(
        self,
        path: Path,
        dsttype: typing.Optional[FileType] = None,
        pipeline: typing.Optional[typing.List[TranscoderName]] = None,
    ) -> typing.Tuple[MultiDiGraph, typing.List[reg.TranscodingPath]]:
        logger.debug(
            "Selecting appropriate pipelines for path:%s dsttype:%s pipeline:%s",
            path,
            dsttype,
            pipeline,
        )
        graph, paths = self.transcoder_registry.map_pipelines(path)
        transcoding_paths = self.pipeline_selector.select_pipelines(
            paths, path, dsttype, pipeline
        )
        return graph, transcoding_paths


class SingleItemSyncer(object):
    def __init__(self, postprocessor: Postprocessor):
        self.postprocessor = postprocessor

    def sync(self, src: Path, dst: Path, transcoding_path: reg.TranscodingPath) -> None:
        logger.debug("Beginning to transcode from %s", src)
        files.ensure_directories_exist([dst.parent.as_posix()])
        in_fn = src.as_posix()
        with files.remover() as tmpfiles:
            for step in transcoding_path.steps:
                prefix = ".tmp-" + step.transcoder_name + dst.stem
                suffix = "." + step.dsttype
                prefix = files.shorten_to_name_max(
                    dst.parent.as_posix(),
                    prefix,
                    8 + len(suffix),
                )
                out_f = tempfile.NamedTemporaryFile(
                    prefix=prefix,
                    dir=dst.parent.as_posix(),
                    suffix=suffix,
                    delete=False,
                )
                out_fn = out_f.name
                tmpfiles.append(out_fn)
                out_f.close()
                logger.debug("Pipeline step: %s", step)
                step.transcode(Path(in_fn), Path(out_fn))
                shutil.copymode(in_fn, out_fn)
                in_fn = out_fn
            self.postprocessor(
                src.as_posix(),
                in_fn,
                transcoding_path.srctype,
                transcoding_path.dsttype,
            )
            os.rename(in_fn, dst)
        logger.debug("Done transcoding to %s", dst)



from pathlib import Path
from typing import Union, List
import unittest

from . import gstreamerffmpeg as mod


class TestGst(unittest.TestCase):
    def test_same_pipeline_as_before(self) -> None:
        src, dst, f = Path("a"), Path("b"), "gst-launch-1.0"
        in_ = [
            "decodebin",
            "audioconvert",
            "audio/x-raw,format=F32LE",
            "wavenc",
        ]
        exp_ = [
            f,
            "-f",
            "giosrc",
            "location=file://%s" % src.absolute().as_posix(),
            "!",
            "decodebin",
            "!",
            "audioconvert",
            "!",
            "audio/x-raw,format=F32LE",
            "!",
            "wavenc",
            "!",
            "filesink",
            "location=%s" % dst.absolute().as_posix(),
        ]
        self.assertListEqual(exp_, mod.gst(src, dst, *in_, force_gst_command=f))

    def test_element_with_parameters(self) -> None:
        src, dst, f = Path("a"), Path("b"), "gst-launch-1.0"
        in_: List[Union[str, List[str]]] = [
            "decodebin",
            "audioconvert",
            [
                "lamemp3enc",
                "encoding-engine-quality=2",
                "quality=0",
            ],
            "xingmux",
        ]
        exp_ = [
            f,
            "-f",
            "giosrc",
            "location=file://%s" % src.absolute().as_posix(),
            "!",
            "decodebin",
            "!",
            "audioconvert",
            "!",
            "lamemp3enc",
            "encoding-engine-quality=2",
            "quality=0",
            "!",
            "xingmux",
            "!",
            "filesink",
            "location=%s" % dst.absolute().as_posix(),
        ]
        self.assertListEqual(exp_, mod.gst(src, dst, *in_, force_gst_command=f))

    def test_candidate_sort_largest_version_wins(self) -> None:
        in_ = [
            "/usr/bin/gst-launch-0.10",
            "/usr/bin/gst-launch-1.0",
            "/usr/bin/gst-launch-1.1",
        ]
        exp = [
            "/usr/bin/gst-launch-1.1",
            "/usr/bin/gst-launch-1.0",
            "/usr/bin/gst-launch-0.10",
        ]
        res = mod.sort_gst_candidates(in_)
        self.assertListEqual(exp, res)

    def test_candidate_sort_unversioned_wins(self) -> None:
        in_ = [
            "/usr/bin/gst-launch",
            "/usr/bin/gst-launch-1.0",
            "/usr/bin/gst-launch-1.1",
        ]
        exp = [
            "/usr/bin/gst-launch",
            "/usr/bin/gst-launch-1.1",
            "/usr/bin/gst-launch-1.0",
        ]
        res = mod.sort_gst_candidates(in_)
        self.assertListEqual(exp, res)



"""
Created on Aug 11, 2012

@author: rudd-o
"""

import collections
import contextlib
import os
from pathlib import Path
import pprint
from queue import Queue
import shutil
import tempfile
import textwrap
import typing
import unittest

from . import core as mod
from ..files import AbsolutePath, Absolutize as A
from ..files import ensure_directories_exist
from ..transcoding import policies, registry, settings, transcoder
from ..transcoding.interfaces import TranscoderName, FileType
from ..transcoding.test_registry import DummyLookup


# FIXME: test musictoolbox.synccli
@contextlib.contextmanager
def synthetic_playlists_fixtures() -> typing.Generator[
    typing.Tuple[
        typing.List[AbsolutePath],
        typing.Dict[
            AbsolutePath,
            typing.List[AbsolutePath],
        ],
    ],
    None,
    None,
]:
    with tempfile.TemporaryDirectory() as d:
        p1, p2 = open(Path(d) / "a.m3u", "w"), open(Path(d) / "b.m3u", "w")
        try:
            p1.write(
                textwrap.dedent(
                    """
                    #test playlist
                    /home/rudd-o/mp3.mp3
                    home/rudd-o/wav.mp3
                    /home/rudd-o/common.mp3
                    """
                )
            )
            p2.write(
                textwrap.dedent(
                    """
                    #test playlist 2
                    /home/rudd-o/playlist2.mp3
                    /home/rudd-o/wav.mp3
                    /home/rudd-o/common.mp3
                    """
                )
            )
        finally:
            p1.close()
            p2.close()

        output = {
            A("/home/rudd-o/mp3.mp3"): [A(p1.name)],
            A(Path(d) / "home/rudd-o/wav.mp3"): [A(p1.name)],
            A("/home/rudd-o/common.mp3"): [A(p1.name), A(p2.name)],
            A("/home/rudd-o/playlist2.mp3"): [A(p2.name)],
            A("/home/rudd-o/wav.mp3"): [A(p2.name)],
        }
        yield [A(p1.name), A(p2.name)], output


@contextlib.contextmanager
def symlinked_playlist_fixtures() -> typing.Generator[
    typing.Tuple[
        typing.List[AbsolutePath],
        typing.Dict[
            AbsolutePath,
            typing.List[AbsolutePath],
        ],
    ],
    None,
    None,
]:
    with tempfile.TemporaryDirectory() as d:
        os.mkdir(Path(d) / "subdir")
        p1 = open(Path(d) / "subdir" / "pl.m3u", "w")
        try:
            p1.write(
                textwrap.dedent(
                    """
                    #test playlist
                    mp3.mp3
                    """
                )
            )
        finally:
            p1.close()
        p2 = Path(d) / "symlink.m3u"
        os.symlink(p1.name, p2)

        output = {
            A(Path(d) / "subdir/mp3.mp3"): [A(p1.name), A(p2)],
        }
        yield [A(p1.name), A(p2)], output


@contextlib.contextmanager
def list_files_recursively_fixtures() -> typing.Generator[
    typing.Tuple[
        str,
        typing.List[AbsolutePath],
        typing.List[typing.Union[int, float]],
    ],
    None,
    None,
]:
    with tempfile.TemporaryDirectory() as d:
        files = [os.path.join(d, str(x)) for x in range(10)]
        e = os.path.join(d, "subdir")
        os.mkdir(e)
        giles = [os.path.join(e, str(x)) for x in range(10)]
        for f in files + giles:
            with open(f, "w"):
                pass
        mtimes = [os.stat(f).st_mtime for f in files + giles]
        yield d, [A(x) for x in files + giles], mtimes


class TestParsePlaylists(unittest.TestCase):
    maxDiff = 65536

    def test_synthetic_playlists(self) -> None:
        with synthetic_playlists_fixtures() as (sources, output):
            files, excs = mod.parse_playlists(sources)
            self.assertEqual(files, output)
            self.assertEqual(excs, [])

    def test_symlinked_playlist(self) -> None:
        with symlinked_playlist_fixtures() as (sources, output):
            files, excs = mod.parse_playlists(sources)
            self.assertEqual(files, output)
            self.assertEqual(excs, [])

    def test_nonexistent_playlists(self) -> None:
        sources = [A("does not exist")]
        files, excs = mod.parse_playlists(sources)
        self.assertEqual(files, {})
        self.assertTrue(excs[0][0], "does not exist")
        self.assertTrue(isinstance(excs[0][1], IOError), excs[0][1])


class TestListFilesRecursively(unittest.TestCase):
    maxDiff = 65536

    def test_simple_case(self) -> None:
        with list_files_recursively_fixtures() as (d, files, _):
            listed_files = mod.list_files_recursively(A(d))
            listed_files.sort()
            shutil.rmtree(d)
            self.assertEqual(files, listed_files)


def syncplaylists_fixtures(
    dd: AbsolutePath,
    fake_songs: typing.List[str],
    fake_playlists: typing.List[typing.List[str]],
) -> typing.List[AbsolutePath]:
    pls: typing.List[AbsolutePath] = []
    for n, c in enumerate(fake_playlists):
        fn = dd / ("%s.m3u" % n)
        pls.append(fn)
        with fn.open("w") as f:
            f.write("\n".join(c))
    for s in fake_songs:
        ss = dd / s
        ensure_directories_exist([ss.parent.as_posix()])
        with ss.open("w") as f:
            f.write("Not a real file")
    return pls


def donothing_postpro(
    unused1: str,
    unused2: str,
    unused3: typing.Optional[str] = None,
    unused4: typing.Optional[str] = None,
) -> None:
    pass


def consume(
    q: Queue[typing.Union[mod.SyncQueueItem, None]]
) -> typing.List[mod.SyncQueueItem]:
    res: typing.List[mod.SyncQueueItem] = []
    while True:
        v = q.get()
        if v is None:
            break
        res.append(v)
    return res


def copypath(ft: str) -> registry.TranscodingPath:
    return registry.TranscodingPath(
        1,
        DummyLookup(),
        [(FileType.by_name(ft), FileType.by_name(ft), TranscoderName("copy"))],
    )


class TestSynchronizer(unittest.TestCase):
    maxDiff = 65535

    def setUp(self) -> None:
        self.td = A(tempfile.mkdtemp(suffix=self.__class__.__name__))

    def tearDown(self) -> None:
        shutil.rmtree(self.td)

    def _makeStack(
        self,
        playlists: typing.List[AbsolutePath],
        target_dir: AbsolutePath,
        forced_policy: typing.Optional[policies.TranscoderPolicy] = None,
        allow_fallback: bool = True,
        force_vfat: bool = False,
    ) -> mod.Synchronizer:
        ts = settings.TranscoderSettings({})
        tr = registry.TranscoderRegistry(ts)
        pipeline = [TranscoderName("copy")]
        policy = (
            forced_policy
            if forced_policy
            else policies.TranscoderPolicy(
                source=None, target=None, transcode_to=None, pipeline=pipeline
            )
        )
        ps = policies.PolicyBasedPipelineSelector(
            policies.TranscoderPolicies([policy]), allow_fallback
        )
        tm = transcoder.TranscodingMapper(tr, ps)
        s = mod.Synchronizer(
            playlists, target_dir, tm, [], donothing_postpro, force_vfat
        )
        return s

    def test_basic(self) -> None:
        in_ = (
            self.td,
            [
                "Albums/Good/A-Ha/Take on me.mp3",
                "Albums/Bad/Ace of Base/Tell your dad?.ogg",
            ],
            [
                ["Albums/Good/A-Ha/Take on me.mp3"],
                [
                    "Albums/Bad/Ace of Base/Tell your dad?.ogg",
                    "Albums/Good/A-Ha/Take on me.mp3",
                ],
            ],
        )
        want = [
            (
                self.td / "Albums/Good/A-Ha/Take on me.mp3",
                self.td / "output/Good/A-Ha/Take on me.mp3",
                copypath("mp3"),
            ),
            (
                self.td / "Albums/Bad/Ace of Base/Tell your dad?.ogg",
                self.td / "output/Bad/Ace of Base/Tell your dad?.ogg",
                copypath("ogg"),
            ),
        ]
        playlists = syncplaylists_fixtures(*in_)
        s = self._makeStack(playlists, self.td / "output")
        got = s.compute_synchronization()
        self.assertEqual(want, got[0])

        q, unused_cancel = s.synchronize(got, 1)
        res = consume(q)
        for _, _, d in res:
            assert not isinstance(d, Exception)
        for _, v, __ in want:
            assert os.path.exists(v), v

    def test_vfat(self) -> None:
        in_ = (
            self.td,
            ["Albums/Bad/Ace of Base/Tell your dad?.ogg"],
            [["Albums/Bad/Ace of Base/Tell your dad?.ogg"]],
        )
        want = [
            (
                self.td / "Albums/Bad/Ace of Base/Tell your dad?.ogg",
                self.td / "output/Tell your dad_.ogg",
                copypath("ogg"),
            ),
        ]
        playlists = syncplaylists_fixtures(*in_)
        s = self._makeStack(playlists, self.td / "output", force_vfat=True)
        got = s.compute_synchronization()
        self.assertEqual(want, got[0])

        q, unused_cancel = s.synchronize(got, 1)
        res = consume(q)
        for _, _, d in res:
            assert not isinstance(d, Exception)
        for _, v, __ in want:
            assert os.path.exists(v), v

    def test_playlist_does_not_exist(self) -> None:
        s = self._makeStack([self.td / "notexist.m3u"], self.td)
        self.assertRaises(FileNotFoundError, s.compute_synchronization)

    def test_no_matching_pipeline(self) -> None:
        in_ = (
            self.td,
            [
                "Albums/Good/A-Ha/Take on me.aardvark",
            ],
            [
                ["Albums/Good/A-Ha/Take on me.aardvark"],
            ],
        )
        want = collections.OrderedDict(
            [
                (
                    self.td / "Albums/Good/A-Ha/Take on me.aardvark",
                    transcoder.NoPipeline(
                        self.td / "Albums/Good/A-Ha/Take on me.aardvark"
                    ),
                ),
            ]
        )
        playlists = syncplaylists_fixtures(*in_)
        s = self._makeStack(
            playlists,
            self.td,
            forced_policy=policies.TranscoderPolicy(
                source=FileType.by_name("aardvark"),
                target=None,
                transcode_to=None,
                pipeline=[TranscoderName("extractaudio")],
            ),
            allow_fallback=False,
        )
        got = s.compute_synchronization()
        self.assertEqual(got[0], [])

        gots, wants = pprint.pformat(got[1]), pprint.pformat(want)
        assert gots == wants

    def test_uppercase_extension(self) -> None:
        in_ = (
            self.td,
            ["Albums/Good/A-Ha/Take on me.OGG"],
            [["Albums/Good/A-Ha/Take on me.OGG"]],
        )
        want = [
            (
                self.td / "Albums/Good/A-Ha/Take on me.OGG",
                self.td / "Take on me.ogg",
                copypath("ogg"),
            ),
        ]
        playlists = syncplaylists_fixtures(*in_)
        s = self._makeStack(playlists, self.td)
        got = s.compute_synchronization()
        self.assertEqual(got[0], want)

    def test_sync_playlists_base(self) -> None:
        in_ = (
            self.td,
            ["Albums/Good/A-Ha/Take on me.OGG"],
            [["Albums/Good/A-Ha/Take on me.OGG"]],
        )
        wantp = """# from: %s\n# was: %s\n%s""" % (
            self.td / "0.m3u",
            "Albums/Good/A-Ha/Take on me.OGG",
            "../Take on me.ogg",
        )
        playlists = syncplaylists_fixtures(*in_)
        s = self._makeStack(playlists, self.td)
        plan = s.compute_synchronization()
        plsync = list(s.synchronize_playlists(plan))
        assert plsync == [
            (self.td / "0.m3u", self.td / "Playlists/0.m3u", None),
        ]
        with (self.td / "Playlists/0.m3u").open("r") as f:
            gotp = f.read()
        self.assertMultiLineEqual(wantp, gotp)

    def test_sync_playlists_symlink(self) -> None:
        in_ = (
            self.td,
            ["Artist/Take on me.ogg"],
            [["Artist/Take on me.ogg"]],
        )
        wantp = """# from: %s\n# was: %s\n%s""" % (
            self.td / "symlinked/symlinked.m3u",
            "Artist/Take on me.ogg",
            "../Take on me.ogg",
        )
        playlists = syncplaylists_fixtures(*in_)
        os.mkdir(self.td / "symlinked")
        os.symlink(self.td / "0.m3u", self.td / "symlinked" / "symlinked.m3u")
        playlists[0] = A(self.td / "symlinked" / "symlinked.m3u")
        s = self._makeStack(playlists, self.td)
        plan = s.compute_synchronization()
        plsync = list(s.synchronize_playlists(plan))
        assert plsync == [
            (
                self.td / "symlinked/symlinked.m3u",
                self.td / "Playlists/symlinked.m3u",
                None,
            ),
        ]
        with (self.td / "Playlists/symlinked.m3u").open("r") as f:
            gotp = f.read()
        self.assertMultiLineEqual(wantp, gotp)

    def test_synchronize_deletions(self) -> None:
        with tempfile.TemporaryDirectory() as d:
            dp = Path(d)
            (dp / "output").mkdir()
            with (dp / "playlist.m3u").open("w") as f:
                f.write("a.mp3")

            with (dp / "a.mp3").open("w") as f:
                f.write("x")

            with (dp / "output/b.mp3").open("w") as f:
                f.write("x")

            (dp / "output/Playlists").mkdir()
            with (dp / "output/Playlists/x.m3u").open("w") as f:
                f.write("x")

            s = self._makeStack([A(dp / "playlist.m3u")], A(dp / "output"))
            plan = s.compute_synchronization()
            want = [
                (A(dp / "output/b.mp3"), None),
                (A(dp / "output/Playlists/x.m3u"), None),
            ]
            got = list(s.synchronize_deletions(plan))
            self.assertListEqual(want, got)



from __future__ import print_function

import logging
import typing

from mutagen.apev2 import APEv2
from mutagen.id3 import ID3
from mutagen import File

logger = logging.getLogger(__name__)


# ======= mp3gain and soundcheck operations ==========

REPLAYGAIN_TAGS = (
    "replaygain_track_gain",
    "replaygain_track_peak",
    "replaygain_album_gain",
    "replaygain_album_peak",
    "replaygain_reference_loudness",
    "RVA2",
)


def printpairs(toprint: list[tuple[str, typing.Any]]) -> None:
    key_tpl = "%%%ds" % max(len(x[0]) for x in toprint)
    for key, tag in toprint:
        print(key_tpl % key, "  ", repr(tag))


def viewmp3norm(files: list[str]) -> None:
    while files:
        file = files.pop(0)

        print()
        print(file)

        try:
            tags = File(file)
        except Exception as e:
            print("No known tags", e)
            tags = None
        try:
            apetags = APEv2(file)
        except Exception as e:
            print("No APE tags", e)
            apetags = None

        toprint: list[tuple[str, typing.Any]] = []
        for key, tag in list(apetags.items()):
            toprint.append((key, tag))
        if toprint:
            print("===APE tags====")
            printpairs(toprint)

        toprint = []
        if tags:
            toprint = []
            for key, tag in list(tags.items()):
                try:
                    if any(
                        [x in tag.desc.lower() or x in key for x in REPLAYGAIN_TAGS]
                    ):
                        toprint.append((key, tag))
                except AttributeError:
                    continue
        if toprint:
            print("===Other tags===")
            printpairs(toprint)


def viewtags(files: typing.List[str]) -> None:
    while files:
        file = files.pop(0)

        print()
        print(file)

        try:
            tags = ID3(file)
        except Exception as e:
            print("No ID3 tags", e)
            tags = None
        try:
            apetags = APEv2(file)
        except Exception as e:
            print("No APE tags", e)
            apetags = None

        toprint: list[tuple[str, typing.Any]] = []
        if apetags:
            print("===APE tags====")
            for key, tag in list(apetags.items()):
                toprint.append((key, tag))
            printpairs(toprint)

        toprint = []
        if tags:
            print("===ID3 tags===")
            for key, tag in list(tags.items()):
                toprint.append((key, tag))
            printpairs(toprint)


# algorithm begins here



import logging
import os
from queue import Queue
from threading import Thread
import typing

import concurrent.futures as fut
import signal as _  # @UnusedImport

from . import algo
from .. import files
from ..files import AbsolutePath, Absolutize
from ..transcoding import registry as reg, transcoder
from ..transcoding.interfaces import Postprocessor
from .interfaces import (
    PathMappingProtocol,
    PathComparisonProtocol,
)


logger = logging.getLogger(__name__)


def parse_playlists(
    sources: typing.List[AbsolutePath],
) -> typing.Tuple[
    typing.Dict[AbsolutePath, typing.List[AbsolutePath]],
    typing.List[typing.Tuple[AbsolutePath, Exception]],
]:
    """
    Take several playlists, and return a dictionary and a list.
    This dictionary is structured as follows:
    - keys: absolute path names mentioned in the playlists
    - values: playlists where the file appeared
    The list is a sequence of (file, Exception) occurred while parsing.
    Symlinked playlists are followed to their targets to extract paths.
    """
    files: typing.Dict[AbsolutePath, typing.List[AbsolutePath]] = {}
    excs: typing.List[typing.Tuple[AbsolutePath, Exception]] = []
    for source in sources:
        try:
            realsource = Absolutize(os.path.realpath(source))
            realsourcedir = realsource.parent
            with open(realsource, "r") as sourcef:
                thisbatch = [
                    Absolutize(os.path.join(realsourcedir, x.strip()))
                    for x in sourcef.readlines()
                    if x.strip() and not x.strip().startswith("#")
                ]
                for path in thisbatch:
                    if path not in files:
                        n: typing.List[AbsolutePath] = []
                        files[path] = n
                    files[path].append(source)
        except Exception as e:
            excs.append((source, e))
    return files, excs


def list_files_recursively(
    directory: AbsolutePath,
) -> typing.List[AbsolutePath]:
    """Return a list of absolute paths from recursively listing a directory"""
    return [
        Absolutize(os.path.join(base, f))
        for base, _, files in os.walk(directory)
        for f in files
    ]


class CallerStopped(Exception):
    pass


SyncQueueItem = typing.Tuple[
    AbsolutePath,
    AbsolutePath,
    typing.Union[
        None,
        Exception,
    ],
]


class SyncPool(Thread):
    def __init__(
        self,
        to_sync: typing.List[
            typing.Tuple[AbsolutePath, AbsolutePath, reg.TranscodingPath]
        ],
        slave: transcoder.SingleItemSyncer,
        max_workers: typing.Optional[int] = None,
    ):
        Thread.__init__(self, daemon=True)
        self.slave = slave
        self.to_sync = to_sync
        self.executor = fut.ThreadPoolExecutor(max_workers=max_workers)
        self.cancelled: typing.List[bool] = []
        self.results: Queue[typing.Union[SyncQueueItem, None]] = Queue()

    def cancel(self) -> None:
        self.cancelled.append(True)
        self.executor.shutdown(wait=True, cancel_futures=True)
        while True:
            # Consume the results that remain.
            r = self.results.get()
            if r is None:
                break

    def run(self) -> None:
        try:
            future_to_url = {
                self.executor.submit(self.slave.sync, s, d, p): (s, d)
                for s, d, p in self.to_sync
            }
            for future in fut.as_completed(future_to_url):
                if self.cancelled:
                    break
                src, dst = future_to_url[future]
                exc: typing.Union[None, Exception] = None
                try:
                    future.result()
                except Exception as e:
                    exc = e
                self.results.put((src, dst, exc))
        finally:
            # FIXME
            # [delete_ignoring_notfound(tmpd) for _, tmpd, d in series]
            self.results.put(None)
            self.executor.shutdown(wait=True, cancel_futures=True)


class Synchronizer(object):
    def __init__(
        self,
        playlists: typing.List[AbsolutePath],
        target_directory: AbsolutePath,
        transcoding_mapper: transcoder.TranscodingMapper,
        exclude_beneath: typing.List[AbsolutePath],
        postprocessor: Postprocessor,
        force_vfat: bool,
    ) -> None:
        self.playlists = playlists
        self.target_directory = target_directory
        self.target_playlist_dir = self.target_directory / "Playlists"

        self.transcoding_mapper = transcoding_mapper
        self.postprocessor: Postprocessor = postprocessor
        self.filesystem_path_mapper = (
            algo.ForceVFATPathMapper(self.target_directory)
            if force_vfat
            else algo.FilesystemPathMapper(self.target_directory)
        )

        self.exclude_beneath = exclude_beneath

    def compute_synchronization(self, unconditional: bool = False) -> algo.SyncRet:
        """Computes synchronization between sources and target."""

        logger.debug("Parsing %s playlists", len(self.playlists))
        source_files, excs = parse_playlists(self.playlists)
        logger.debug("Discovered %s source files", len(source_files))
        if excs:
            for pl, e in excs:
                logger.error("Cannot scan playlist %s: %s", pl, e)
            raise e

        logger.debug("Scanning target directory %s", self.target_directory)
        try:
            target_files = list_files_recursively(self.target_directory)
            logger.debug("Discovered %s target files", len(target_files))
        except Exception as e:
            logger.error("Cannot scan target directory: %s", e)
            raise e

        source_basedir = Absolutize(
            os.path.commonprefix([s.parent for s in source_files])
        )

        class AbsoluteMapperAdapter(object):
            def __init__(self, mapper: transcoder.TranscodingMapper):
                self.mapper = mapper

            def map(self, src: AbsolutePath) -> AbsolutePath:
                return Absolutize(self.mapper.map(src))

        source_mappers: typing.List[PathMappingProtocol] = [
            AbsoluteMapperAdapter(self.transcoding_mapper),
        ]
        target_mappers: typing.List[PathMappingProtocol] = [self.filesystem_path_mapper]
        transcode_pather = self.transcoding_mapper

        comparator: PathComparisonProtocol = algo.ModTimestampComparer()
        if unconditional:
            comparator = algo.SourceAlwaysNewer()

        # Also exclude from deletion all playlists that will be synced.
        exclude_beneath = self.exclude_beneath + [
            self.target_playlist_dir / p.name for p in self.playlists
        ]

        return algo.compute_synchronization(
            list(source_files),
            source_basedir,
            target_files,
            self.target_directory,
            source_mappers,
            target_mappers,
            transcode_pather,
            comparator,
            exclude_beneath,
        )

    def synchronize(
        self,
        sync_plan: algo.SyncRet,
        concurrency: int,
    ) -> typing.Tuple[
        Queue[typing.Union[SyncQueueItem, None]],
        typing.Callable[[], None],
    ]:
        """
        Computes synchronization between sources and target, then
        gets ready to sync using a thread pool, dispatching tasks to it.
        The start of the sync process happens immediately, and synchronization
        happens asynchronously.

        Returns a tuple of Queue with result objects, and a callable that
        can be called to cancel the process.  Once the cancel callable
        returns, the process has stopped.  The result object is a tuple
        with (src, dst, optional Exception).
        """
        to_sync, _, __, ___ = sync_plan
        max_workers: typing.Optional[int] = (
            concurrency if (concurrency and concurrency > 0) else None
        )

        logger.info(
            "Synchronizing %s items with %s threads",
            len(to_sync),
            max_workers if max_workers else "automatic number of",
        )
        slave = transcoder.SingleItemSyncer(self.postprocessor)
        t = SyncPool(to_sync, slave, max_workers=max_workers)
        t.start()

        return t.results, t.cancel

    def synchronize_playlists(
        self, sync_plan: algo.SyncRet, dryrun: bool = False
    ) -> typing.Generator[
        typing.Tuple[AbsolutePath, AbsolutePath, typing.Union[Exception, None]],
        None,
        None,
    ]:
        """
        Once synchronization of files has been done, synchronization of
        playlists is possible.

        The entry conditions for this function are the same as the entry
        conditions of synchronize().

        It yields tuples of (source_path, target_path, optional Exception)
        where the exception will not be None if there was a problem
        synchronizing a particular playlist.

        Symlinked playlists are followed to their target files to compute
        their synchronization.
        """
        # FIXME if adding params to the following call, add them above too
        will_sync_list, wont_sync, already_synced, _ = sync_plan
        will_sync = dict((s, d) for s, d, _ in will_sync_list)

        if not dryrun:
            try:
                files.ensure_directories_exist([self.target_playlist_dir.as_posix()])
            except Exception as e:
                yield (self.target_playlist_dir, self.target_playlist_dir, e)
                return

        for p in self.playlists:
            newp = self.target_playlist_dir / p.name
            oldp, p = p, Absolutize(os.path.realpath(p))
            try:
                pdir = p.parent
                with p.open("r") as pf:
                    pfl = pf.readlines()
                newpfl = []
                newpfl.append("# from: %s\n" % oldp)
                for l in pfl:
                    if l.startswith("#") or not l.strip():
                        newpfl.append(l)
                        continue
                    cr = l.endswith("\n")
                    orgl = "# was: " + l.strip() + "\n"
                    newpfl.append(orgl)
                    l = l.strip()
                    truel = Absolutize(os.path.join(pdir, l))
                    if truel in will_sync:
                        l = will_sync[truel].as_posix()
                        l = os.path.relpath(l, self.target_playlist_dir.as_posix())
                    elif truel in already_synced:
                        l = already_synced[truel].as_posix()
                        l = os.path.relpath(l, self.target_playlist_dir.as_posix())
                    elif truel in wont_sync:
                        l = "# not synced because of %s" % wont_sync[truel]
                    else:
                        assert 0, (l, truel)
                    if cr:
                        l += "\n"
                    newpfl.append(l)
                sync = True
                try:
                    with newp.open("r") as oldpfl:
                        if oldpfl.read() == "".join(newpfl):
                            sync = False
                except FileNotFoundError:
                    pass
                if sync:
                    if not dryrun:
                        with newp.open("w") as newpf:
                            newpf.writelines(newpfl)
                            newpf.flush()
                    yield (oldp, newp, None)
            except Exception as e:
                yield (oldp, newp, e)

    def synchronize_deletions(
        self, sync_plan: algo.SyncRet, dryrun: bool = False
    ) -> typing.Generator[
        typing.Tuple[AbsolutePath, typing.Union[Exception, None]], None, None
    ]:
        """
        Once synchronization of files and playlists has been done, sync
        of deletions is possible.

        The entry conditions for this function are the same as the entry
        conditions of synchronize_playlists().

        It yields tuples of (path, optional Exception) where the exception
        will not be None if there was a problem deleting a particular file.
        """
        _, __, ___, deleting = sync_plan

        for t in deleting:
            try:
                if not dryrun:
                    t.unlink(True)
                yield (t, None)
            except Exception as e:
                yield (t, e)


# =================== end synchronizer code ========================



import sys
import typing
from mutagen.apev2 import APEv2, APENoHeaderError


def detect_broken_ape_tags(files: typing.List[str]) -> None:
    while files:
        f = files.pop(0)
        try:
            APEv2(f)
        except APENoHeaderError:
            pass
        except KeyError as e:
            print(f)
        except Exception as e:
            print("while processing %r: %s" % (f, e), file=sys.stderr)


def detect_missing_ape_tags(files: typing.List[str]) -> None:
    while files:
        f = files.pop(0)
        try:
            APEv2(f)
        except APENoHeaderError:
            print(f)
        except KeyError as e:
            pass
        except Exception as e:
            print("while processing %r: %s" % (f, e), file=sys.stderr)



import os
import typing
import contextlib
from pathlib import Path
from threading import Lock


AbsolutePath = typing.NewType("AbsolutePath", Path)


def Absolutize(p: typing.Union[Path, str]) -> AbsolutePath:
    if isinstance(p, Path):
        p = p.as_posix()
    pp: AbsolutePath = Path(os.path.abspath(p))  # type: ignore
    return pp


_dir_lock = Lock()


def ensure_directories_exist(dirs: typing.List[str]) -> None:
    global _dir_lock
    with _dir_lock:
        for t in dirs:
            if not t:
                continue
            if not os.path.exists(t):
                os.makedirs(t)


def ensure_files_gone(files: typing.List[str]) -> None:
    for t in files:
        if not t:
            continue
        try:
            os.unlink(t)
        except FileNotFoundError:
            pass


@contextlib.contextmanager
def remover() -> typing.Generator[typing.List[str], None, None]:
    """
    Yields an empty list of files.  At the end of the context,
    removes all files in the list of files.
    """
    files: typing.List[str] = []
    try:
        yield files
    finally:
        ensure_files_gone(files)


def shorten_to_name_max(directory: str, name: str, strip_extra_chars: int) -> str:
    pathconf_directory = os.path.abspath(directory)
    while not os.path.isdir(pathconf_directory):
        pathconf_directory = os.path.dirname(pathconf_directory)
    maxfilenamelen = os.pathconf(pathconf_directory, "PC_NAME_MAX")
    name = name[: maxfilenamelen - strip_extra_chars]
    return name


def all_files(paths: list[str], recursive: bool = True) -> list[str]:
    allfiles: list[str] = []
    for f in paths:
        if os.path.isdir(f):
            if recursive:
                files = [
                    os.path.join(root, x) for root, _, fs in os.walk(f) for x in fs
                ]
            else:
                files = [
                    os.path.join(f, x)
                    for x in os.listdir(f)
                    if not os.path.isdir(os.path.join(f, x))
                ]
        else:
            files = [f]
        allfiles.extend(files)
    return allfiles



import argparse
import glob
import logging
import os
import sys
import typing

import mutagen

from ..logging import basicConfig


def sort_by_track_number(paths: typing.List[str]) -> typing.List[str]:
    metadatas = [(f, mutagen.File(f)) for f in paths]
    relevant = []
    relevant.extend(
        [(int(m["tracknumber"][0]), f, m) for f, m in metadatas if "tracknumber" in m]
    )
    relevant.extend(
        [
            (int(m["TRCK"].text[0].split("/")[0]), f, m)
            for f, m in metadatas
            if "TRCK" in m
        ]
    )
    relevant.extend(
        [(99, f, m) for f, m in metadatas if f not in [b for _, b, __ in relevant]]
    )
    relevant = list(sorted(relevant))
    return [f for _, f, __ in relevant]


def make_album_playlist(
    playlistpath: str, paths: typing.List[str], dryrun: bool = False
) -> None:
    playlist_dir = os.path.dirname(playlistpath)
    files: typing.List[str] = []
    for p in paths:
        if os.path.isdir(p):
            for spec in ["*.mp3", "*.ogg", "*.mpc", "*.flac"]:
                files.extend(glob.iglob(os.path.join(p, spec)))
        else:
            files.append(p)
    sortedfiles = sort_by_track_number(files)
    relativefiles = [os.path.relpath(f, playlist_dir) for f in sortedfiles]
    if dryrun:
        playlist_fileobj = sys.stdout
    else:
        playlist_fileobj = open(playlistpath, "w")
    playlist_fileobj.write("\n".join(relativefiles))


def get_parser() -> argparse.ArgumentParser:
    program_name = os.path.basename(sys.argv[0])
    program_shortdesc = (
        "Creates a playlist from passed directories and files, sorted by track number"
    )

    parser = argparse.ArgumentParser(
        prog=program_name,
        description=program_shortdesc,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-n",
        "--dry-run",
        dest="dryrun",
        action="store_true",
        help="do nothing except show what will happen [default: %(default)s]",
    )
    parser.add_argument(
        dest="paths", help="files and directories to include in playlist", nargs="+"
    )
    parser.add_argument(
        dest="playlist", help="path to an M3U playlist to write", nargs=1
    )
    return parser


def main() -> int:
    basicConfig(main_module_name=__name__, level=logging.WARNING)

    parser = get_parser()
    args = parser.parse_args()

    make_album_playlist(args.playlist[0], args.paths, dryrun=args.dryrun)
    return 0



from pathlib import Path
from typing import List, Tuple
import unittest

from . import config as cfg
from . import registry as reg
from . import settings as set
from .codecs.base import NoSettings
from .interfaces import FileType, TranscoderName, TranscoderProtocol


class DummyTranscoder(NoSettings):
    cost = 0

    def transcode(self, src: Path, dst: Path) -> None:  # @UnusedVariable
        return None

    def can_transcode(self, src: Path) -> List[FileType]:  # @UnusedVariable
        return []


class DummyLookup(object):
    def get_transcoder(
        self, transcoder_name: TranscoderName  # @UnusedVariable
    ) -> TranscoderProtocol:
        return DummyTranscoder({})


def mp(cost: int, steps: List[Tuple[str, str, str]]) -> reg.TranscodingPath:
    return reg.TranscodingPath(
        cost,
        DummyLookup(),
        [
            (FileType.by_name(x[0]), FileType.by_name(x[1]), TranscoderName(x[2]))
            for x in steps
        ],
    )


class TestTranscoderRegistry(unittest.TestCase):
    def setUp(self) -> None:
        self.r = reg.TranscoderRegistry(cfg.DefaultTranscoderConfiguration.settings)

    def test_basic_map(self) -> None:
        have = "a.mp3"
        exp = [
            mp(1, [("mp3", "mp3", "copy")]),
            mp(10, [("mp3", "wav", "audiotowav")]),
            mp(20, [("mp3", "wav", "audiotowav"), ("wav", "opus", "wavtoopus")]),
            mp(20, [("mp3", "wav", "audiotowav"), ("wav", "ogg", "wavtoogg")]),
        ]
        unused_graph, got = self.r.map_pipelines(Path(have))
        self.assertListEqual([str(s) for s in exp], [str(s) for s in got])

    def test_copy_is_first(self) -> None:
        have = "a.mp3"
        exp = [mp(1, [("mp3", "mp3", "copy")])]
        unused_graph, got = self.r.map_pipelines(Path(have))
        self.assertListEqual([str(s) for s in exp[0:1]], [str(s) for s in got[0:1]])

    def test_transcoder_settings_for_unknown_transcoder(self) -> None:
        c = set.TranscoderSettings({"unknown": {"a": "b"}})
        self.assertRaises(ValueError, reg.TranscoderRegistry, c)



"""
Created on Aug 11, 2012

@author: rudd-o
"""

import collections
import logging
import os
import pathlib
import typing

import psutil  # type: ignore

from ..files import AbsolutePath, Absolutize
from ..transcoding.registry import (
    TranscodingPathLookupProtocol,
    TranscodingPath,
)
from .interfaces import (
    PathMappingProtocol,
    PathComparisonProtocol,
)


logger = logging.getLogger(__name__)


def within(path: AbsolutePath, subpath: AbsolutePath) -> bool:
    """Return True if path equals subpath or subpath is within path."""
    if path == subpath:
        return True
    return subpath.is_relative_to(path)


def delete_ignoring_notfound(f: pathlib.Path) -> None:
    try:
        os.unlink(f.as_posix())
    except FileNotFoundError:
        pass


def vfatprotect(f: str) -> str:
    """Replace illegal characters in VFAT file system paths."""
    for illegal in '?<>\\:*|"^':
        f = f.replace(illegal, "_")
    while "./" in f:
        f = f.replace("./", "/")
    while " /" in f:
        f = f.replace(" /", "/")
    return f


def get_mptypes() -> typing.Dict[str, str]:
    """Return a mapping of mount points and their file system types."""
    return dict([(p.mountpoint, p.fstype) for p in psutil.disk_partitions()])


def get_fstype(
    p: AbsolutePath, mptypes: typing.Dict[str, str]
) -> typing.Tuple[str, AbsolutePath]:
    """Return the file system type corresponding to a path, along with the file system's mount point."""
    assert os.name != "nt"
    for deepest_mountpoint in p.parents:
        if deepest_mountpoint.as_posix() in mptypes:
            break
    return mptypes[deepest_mountpoint.as_posix()], deepest_mountpoint


class FilesystemPathMapper(object):
    def __init__(self, target_dir: AbsolutePath) -> None:  # @UnusedVariable
        self.mptypes = get_mptypes()
        self.paths_seen: typing.Dict[AbsolutePath, AbsolutePath] = {}

    def map(self, path: AbsolutePath) -> AbsolutePath:
        fstype, deepest_mountpoint = get_fstype(path, self.mptypes)
        if fstype not in ("vfat", "ntfs"):
            return path

        head, tail = deepest_mountpoint, path.relative_to(deepest_mountpoint)
        tailstr = pathlib.Path(vfatprotect(tail.as_posix()))
        tailstr_lower = pathlib.Path(tailstr.as_posix().lower())
        normalized_path = Absolutize(head / tailstr)
        normalized_path_lower = Absolutize(head / tailstr_lower)

        if normalized_path_lower not in self.paths_seen:
            self.paths_seen[normalized_path_lower] = normalized_path
        return self.paths_seen[normalized_path_lower]


class ForceVFATPathMapper(FilesystemPathMapper):
    def __init__(self, target_dir: AbsolutePath) -> None:
        FilesystemPathMapper.__init__(self, target_dir)
        self.mptypes = {target_dir.as_posix(): "vfat", "/": self.mptypes["/"]}


def vfatcompare(s: typing.Union[int, float], t: typing.Union[int, float]) -> int:
    """Compare two VFAT timestamps — if they are close enough, return "same"."""
    x = int(s) - int(t)
    if x >= 2:
        return 1
    elif x <= -2:
        return -1
    return 0


class SourceAlwaysNewer(object):
    def compare(self, __arg1: AbsolutePath, __arg2: AbsolutePath) -> int:
        return 1


class ModTimestampComparer(object):
    def __init__(self) -> None:
        self.mptypes = get_mptypes()

    def compare(self, path1: AbsolutePath, path2: AbsolutePath) -> int:
        fstypes = set(get_fstype(p, self.mptypes)[0] for p in [path1, path2])
        if "vfat" in fstypes:
            comparator = vfatcompare
        else:
            comparator = lambda x, y: (1 if x > y else (-1 if x < y else 0))

        try:
            st2 = path2.stat()
        except FileNotFoundError:
            # The target file does not exist, so we always return the
            # source file as newer.
            return 1
        st1 = path1.stat()

        try:
            t1 = st1.st_mtime_ns / 1000000000
            t2 = st2.st_mtime_ns / 1000000000
        except AttributeError:
            t1 = st1.st_mtime
            t2 = st2.st_mtime

        return comparator(t1, t2)


class SyncError(Exception):
    pass


class Conflict(SyncError):
    def __init__(
        self, source: AbsolutePath, target: AbsolutePath, predecessor: AbsolutePath
    ):
        self.source = source
        self.target = target
        self.predecessor = predecessor

    def __str__(self) -> str:
        return (
            "<Conflict: file %s cannot be synced to %s because %s was already synced there>"
            % (self.source, self.target, self.predecessor)
        )

    def __eq__(self, other: typing.Any) -> bool:
        return str(self) == str(other)


SyncRet = typing.Tuple[
    typing.List[typing.Tuple[AbsolutePath, AbsolutePath, TranscodingPath]],
    typing.Dict[AbsolutePath, Exception],
    typing.Dict[AbsolutePath, AbsolutePath],
    typing.List[AbsolutePath],
]


def compute_synchronization(
    source_files: typing.List[AbsolutePath],
    source_basedir: AbsolutePath,
    target_files: typing.List[AbsolutePath],
    target_basedir: AbsolutePath,
    source_mappers: typing.List[PathMappingProtocol],
    target_mappers: typing.List[PathMappingProtocol],
    transcode_pather: TranscodingPathLookupProtocol,
    comparator: PathComparisonProtocol,
    exclude_beneath: typing.Optional[typing.List[AbsolutePath]] = None,
) -> SyncRet:
    """
    Compute a synchronization schedule based on a dictionary of
    {source_filename:mtime} and a dictionary of {target_filename:mtime}.
    Paths must be absolute or the behavior of this function is undefined.
    All paths must be contained in their respective specified directories.
    mtime in the source dictionary can be an exception, in which case it will
    be presumed that the source file in question cannot be transferred.

    The synchronization aims to discover the target file names based on
    the source file names, and then automatically figure out which files
    need to be transferred, based on their absence in the remote site,
    or their modification date.

    This function accepts a path mapping callable that will transform
    a partial source file name, and the full path to the source file name,
    into the desired partial target file name prior to performing the
    date comparison.  By default, it is an identity function x, y => x.

    This function also accepts a time comparator function that will
    be used to perform the comparison between source and target
    modification times.  The comparator is the standard 1 0 -1 comparator
    that takes x,y where it returns 1 if x > y, 0 if x and y are identical,
    and -1 if x < y.  The comparator gets passed the source mtime as the
    first parameter, and the target mtime as the second parameter.
    FAT file system users may want to pass a custom comparator that takes
    into consideration the time resolution of FAT32 file systems
    (greater than 2 seconds).

    Return four values in a tuple:
        1. A dictionary {s:t} where s is the source file name, and
           t is the desired target file name after transfer.
        2. A dictionary {s:e} where s is the source file name, and
           e is the exception explaining why it cannot be synced,
        3. A dictionary {s:e} of files that will be skipped, with their
           corresponding would-be targets that already were transferred,
        4. A list of files that will be deleted from the destination.
    """
    exclude_beneath = exclude_beneath or []

    will_transfer: typing.List[
        typing.Tuple[AbsolutePath, AbsolutePath, TranscodingPath]
    ] = []
    cant_transfer = collections.OrderedDict()
    already_transferred = collections.OrderedDict()
    deleting = collections.OrderedDict()

    def multimap(
        f: AbsolutePath, mappers: typing.List[PathMappingProtocol]
    ) -> AbsolutePath:
        for mapper in mappers:
            f = Absolutize(mapper.map(f))
        return f

    for t in target_files:
        if any(within(exclude_root, t) for exclude_root in exclude_beneath):
            # Do not delete any files within the exclude dirs.
            pass
        else:
            # This file does not reside in the exclude dir.
            # We default to deleting it, until such time that we later
            # decide it must be kept back.
            deleting[t] = True
        # Warm up the target mappers so that paths with different casing
        # will not have a problem later on being discovered.
        multimap(t, target_mappers)

    already_processed: typing.Dict[AbsolutePath, bool] = {}
    already_foreseen: typing.Dict[AbsolutePath, AbsolutePath] = {}

    for src in source_files:
        if not within(source_basedir, src):
            raise ValueError(
                "source path %r not within source dir %r" % (src, source_basedir)
            )

        if src in already_processed:
            continue
        already_processed[src] = True

        try:
            src_mapped = multimap(src, source_mappers)
            tpath = transcode_pather.lookup(src)[0]
        except Exception as e:
            cant_transfer[src] = e
            continue

        rel = src_mapped.relative_to(source_basedir)
        absp = target_basedir / rel

        tgt = multimap(absp, target_mappers)

        if any(within(exclude_root, tgt) for exclude_root in exclude_beneath):
            # Do not sync any files within the exclude dirs.
            continue

        try:
            if tgt in already_foreseen:
                raise Conflict(src, tgt, already_foreseen[tgt])
            if comparator.compare(src, tgt) > 0:
                will_transfer.append((src, tgt, tpath))
                already_foreseen[tgt] = src
            else:
                already_transferred[src] = tgt
        except Exception as e:
            cant_transfer[src] = e
        deleting[tgt] = False

    return (
        will_transfer,
        cant_transfer,
        already_transferred,
        [d for d, y in deleting.items() if y],
    )



import logging
from pathlib import Path
from typing import List, Optional

from . import registry
from .interfaces import FileType, TranscoderName


logger = logging.getLogger(__name__)


def select_pipelines(
    transcoding_paths: List[registry.TranscodingPath],
    src: Path,
    dsttypes: List[FileType],
    pipeline: Optional[List[TranscoderName]] = None,
) -> List[registry.TranscodingPath]:
    srctype = FileType.by_name(FileType.from_path(src))
    # First, we filter all the paths whose source type does not match the requested source type.
    transcoding_paths = [
        p for p in transcoding_paths if (p.steps[0].srctype == srctype)
    ]
    # Then, we filter all the paths whose intermediate or final destination type do not match
    # the requested destination type.
    if dsttypes:
        matches = []
        for p in transcoding_paths:
            chain = [step.srctype for step in p.steps] + [p.steps[-1].dsttype]
            if dsttypes[-1] == chain[-1]:
                # Final step matches.
                if len(dsttypes) == 1:
                    matches.append(p)
                else:
                    if dsttypes[0] in chain[:-1]:
                        matches.append(p)
        transcoding_paths = matches
    # Then, if the user requested a specific pipeline, we filter for it here.
    if pipeline is not None and len(pipeline) > 0:
        transcoding_paths = [
            p
            for p in transcoding_paths
            if len(p.steps) == len(pipeline)
            and all(
                transcoder_name == step.transcoder_name
                for transcoder_name, step in zip(pipeline, p.steps)
            )
        ]
    return transcoding_paths


class TranscoderPolicy(object):
    def __init__(
        self,
        source: Optional[FileType],
        target: Optional[FileType],
        transcode_to: Optional[FileType],
        pipeline: Optional[List[TranscoderName]],
    ):
        self.source = source
        self.target = target
        self.transcode_to = transcode_to
        self.pipeline = pipeline if pipeline else []

    def match(
        self,
        srctype: Optional[FileType] = None,
        dsttype: Optional[FileType] = None,
    ) -> bool:
        wildcard = FileType.by_name("*")
        match_src = (
            not self.source
            or self.source == wildcard
            or not srctype
            or self.source == srctype
        )
        match_dst = (
            not self.target
            or self.target == wildcard
            or not dsttype
            or self.target == dsttype
        )
        return match_src and match_dst

    def __str__(self) -> str:
        parts = []
        if self.source:
            parts.append("source: %s" % self.source)
        if self.target:
            parts.append("target: %s" % self.target)
        if self.pipeline:
            parts.append("pipeline: %s" % " | ".join(self.pipeline))
        if self.transcode_to:
            parts.append("transcode_to: %s" % self.transcode_to)
        return "[" + ", ".join(parts) + "]"


FallbackPolicy = TranscoderPolicy(FileType.by_name("*"), None, None, [])


class TranscoderPolicies(object):
    def __init__(self, policies: List[TranscoderPolicy]):
        self.policies = policies

    def get_policies_for(
        self,
        srctype: Optional[FileType] = None,
        dsttype: Optional[FileType] = None,
    ) -> List[TranscoderPolicy]:
        t = [p for p in self.policies if p.match(srctype, dsttype)]
        return t

    def __str__(self) -> str:
        return "<" + ", ".join("%s" % p for p in self.policies) + ">"


class PolicyBasedPipelineSelector(object):
    def __init__(self, policies: TranscoderPolicies, allow_fallback: bool = True):
        self.policies = policies
        self.allow_fallback = allow_fallback

    def select_pipelines(
        self,
        transcoding_paths: List[registry.TranscodingPath],
        src: Path,
        dsttype: Optional[FileType] = None,
        pipeline: Optional[List[TranscoderName]] = None,
    ) -> List[registry.TranscodingPath]:
        srctype = FileType.from_path(src)
        # Get all policies applicable for the combo of srctype and possibly dsttype.
        policies = self.policies.get_policies_for(srctype, dsttype)
        if self.allow_fallback:
            # Add a fallback policy that allows the system to fall back in case
            # none of the matching policies actually produce at least pipeline.
            policies += [FallbackPolicy]
        for policy in policies:
            # Destination type selection for pipeline lookup.
            # Legend:
            #
            # transcode_to:
            #   The policy specified a specific final format to transcode to.
            #   This is useful when the user knows a particular file format
            #   would normally transcode to another one, but wants to
            #   further transcode from that other one to a final format.
            # target:
            #   The policy specified a target format, usually as match.
            if policy.transcode_to and policy.target:
                dsttypes = [policy.target, policy.transcode_to]
            elif policy.transcode_to:
                dsttypes = [policy.transcode_to]
            elif policy.target:
                dsttypes = [policy.target]
            else:
                dsttypes = []

            limit_to_pipeline = (
                pipeline if pipeline else (policy.pipeline if policy.pipeline else None)
            )
            transcoding_paths_for_policy = select_pipelines(
                transcoding_paths, src, dsttypes, limit_to_pipeline
            )
            if transcoding_paths_for_policy:
                return transcoding_paths_for_policy
        return []


def NoPolicyPipelineSelector() -> PolicyBasedPipelineSelector:
    return PolicyBasedPipelineSelector(TranscoderPolicies([]))



import io
import logging
import os
from typing import TextIO, Optional, List

import xdg.BaseDirectory
import yaml  # type: ignore

from .interfaces import TranscoderName, FileType
from .policies import (
    TranscoderPolicies,
    TranscoderPolicy,
)
from .settings import TranscoderSettings


# Sample policy file:
sample_policy_file = """
policies:
- source: abc
  target: def
settings:
#  copy: {}
# The copy transcoder supports no settings.
#  another:
#     abc: def
"""


logger = logging.getLogger(__name__)


class TranscoderConfiguration(object):

    policies: TranscoderPolicies = TranscoderPolicies([])
    settings: TranscoderSettings = TranscoderSettings({})

    def __str__(self) -> str:
        return "<TranscoderConfiguration policies: %s  settings: %s>" % (
            self.policies,
            self.settings,
        )


DefaultTranscoderConfiguration = TranscoderConfiguration()


class TranscoderConfigurationLoader(yaml.SafeLoader):  # type: ignore
    def construct_transcoder_policy(
        self, node: yaml.MappingNode
    ) -> Optional[TranscoderPolicy]:
        if not node.tag.endswith(":map"):
            raise ValueError(
                "a transcoder policy must be a dictionary of policy settings"
            )
        source: Optional[FileType] = None
        target: Optional[FileType] = None
        transcode_to: Optional[FileType] = None
        pipeline: List[TranscoderName] = []
        for key, val in node.value:
            if key.value == "source":
                v = self.construct_scalar(val)
                if not isinstance(v, str):
                    raise ValueError(
                        "a transcoder source must be a file type in string form"
                    )
                source = FileType.by_name(v)
            elif key.value == "target":
                v = self.construct_scalar(val)
                if not isinstance(v, str):
                    raise ValueError(
                        "a transcoder target must be a file type in string form"
                    )
                target = FileType.by_name(v)
            elif key.value == "transcode_to":
                v = self.construct_scalar(val)
                if not isinstance(v, str):
                    raise ValueError(
                        "a transcoder transcode_to value must be a file type in string form"
                    )
                transcode_to = FileType.by_name(v)
            elif key.value == "pipeline":
                v: list[str] = self.construct_sequence(val)  # type: ignore
                if not isinstance(v, list):
                    raise ValueError(
                        "a transcoder pipeline must be a list of transcoder names"
                    )
                pipeline = [TranscoderName(x) for x in v]
            else:
                raise ValueError(
                    "transcoder policies do not know setting %r" % key.value
                )
        if source or target or transcode_to or pipeline:
            return TranscoderPolicy(
                source=source,
                target=target,
                transcode_to=transcode_to,
                pipeline=pipeline,
            )
        return None

    def construct_transcoder_policies(
        self, node: yaml.SequenceNode
    ) -> TranscoderPolicies:
        if not node.tag.endswith(":seq"):
            raise ValueError("transcoder policies must be a list of policies")
        ps = []
        for v in node.value:
            vv = self.construct_transcoder_policy(v)
            if vv is not None:
                ps.append(vv)
        return TranscoderPolicies(ps)

    def construct_transcoder_settings(
        self, node: yaml.MappingNode
    ) -> TranscoderSettings:
        if not node.tag.endswith(":map"):
            raise ValueError(
                "transcoder settings must be a dictionary of {transcoder name: settings{}}"
            )
        v = yaml.SafeLoader.construct_mapping(self, node, deep=True)
        return TranscoderSettings(v)

    def construct_document(self, node):  # type: ignore
        cfg = TranscoderConfiguration()
        for unused_n, (key, val) in enumerate(node.value):
            if key.value == "policies":
                cfg.policies = self.construct_transcoder_policies(val)
            elif key.value == "settings":
                cfg.settings = self.construct_transcoder_settings(val)
            else:
                raise ValueError("%r is not permitted in the configuration" % key.value)
        return cfg

    @classmethod
    def from_file(cls, fobject: TextIO) -> TranscoderConfiguration:
        p = yaml.load(fobject, Loader=cls)
        if p is None:
            return DefaultTranscoderConfiguration
        assert isinstance(p, TranscoderConfiguration), p
        return p


def transcoding_config_default_filename() -> str:
    return os.path.join("musictoolbox", "transcoding.yaml")


def transcoding_config_default_path() -> str:
    return os.path.join(
        xdg.BaseDirectory.xdg_config_home, transcoding_config_default_filename()
    )


def load_transcoding_config(p: Optional[str] = None) -> TranscoderConfiguration:
    """
    Config file none-> load first found config.
    Config file empty string -> do not load any config.
    """
    if p == "":
        logger.debug(
            "User specified empty config file; explicitly skipping any config file and using defaults"
        )
        return DefaultTranscoderConfiguration

    logger.debug("Specified configuration file: %s", p)
    p = p or xdg.BaseDirectory.load_first_config(transcoding_config_default_filename())
    logger.debug("Actually discovered configuration file: %s", p)
    if p:
        logger.debug("Configuration file %s exists; loading it", p)
        with open(p, "r") as f:
            cfg = TranscoderConfigurationLoader.from_file(f)
            logger.debug("Loaded configuration: %s", cfg)
            return cfg

    logger.debug("No configuration file %s; using defaults", p)
    return DefaultTranscoderConfiguration


if __name__ == "__main__":
    with io.StringIO(sample_policy_file) as f:
        cfg = TranscoderConfigurationLoader.from_file(f)
        print(cfg)



import collections
import logging
import os
from pathlib import Path
from typing import List, Dict, Tuple, Set, Type, Protocol, Any

import pkg_resources  # type: ignore

import networkx as nx  # type: ignore

from .interfaces import (
    FileType,
    TranscoderName,
    TranscoderProtocol,
    TranscoderLookupProtocol,
)
from .settings import TranscoderSettings


logger = logging.getLogger(__name__)


class TranscodingStep(object):
    def __init__(
        self,
        transcoder_db: TranscoderLookupProtocol,
        srctype: FileType,
        dsttype: FileType,
        transcoder_name: TranscoderName,
    ):
        self.srctype = srctype
        self.dsttype = dsttype
        self.transcoder_name = transcoder_name
        self.transcoder_db = transcoder_db

    def __str__(self) -> str:
        return "%s --(%s)--> %s" % (self.srctype, self.transcoder_name, self.dsttype)

    def __repr__(self) -> str:
        return self.__str__()

    def transcode(self, src: Path, dst: Path) -> None:
        transcoder = self.transcoder_db.get_transcoder(self.transcoder_name)
        return transcoder.transcode(src, dst)


class TranscodingPath(object):
    cost: int = -1
    steps: List[TranscodingStep] = []

    def __init__(
        self,
        cost: int,
        transcoder_db: TranscoderLookupProtocol,
        steps_list: List[Tuple[FileType, FileType, TranscoderName]],
    ):
        self.cost = cost
        self.steps = []
        for step in steps_list:
            self.steps.append(TranscodingStep(transcoder_db, *step))

    def __str__(self) -> str:
        return "< %s >" % " | ".join(str(s) for s in self.steps)

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other: Any) -> bool:
        return str(self) == str(other)

    @property
    def srctype(self) -> FileType:
        return self.steps[0].srctype

    @property
    def dsttype(self) -> FileType:
        return self.steps[-1].dsttype


class TranscodingPathLookupProtocol(Protocol):
    def lookup(self, __arg: Path) -> List[TranscodingPath]:
        pass


class TranscoderRegistry(object):

    loaded_entry_points = False
    transcoder_factories: Set[Type[TranscoderProtocol]] = set()
    transcoders: Dict[TranscoderName, TranscoderProtocol] = {}

    def __init__(self, transcoder_settings: TranscoderSettings):
        # FIXME: the following import should be done with a zope interfaces registry,
        # and the Zope component architecture, since  the classes have already been
        # registered as TranscoderProtocol.
        from . import codecs  # @UnusedImport

        if not self.__class__.loaded_entry_points:
            for ep in pkg_resources.iter_entry_points(
                group="musictoolbox.transcoding.codecs"
            ):
                try:
                    factory = ep.load()
                    self.__class__.transcoder_factories.add(factory)
                    self.transcoder_factories.add(factory)
                except pkg_resources.DistributionNotFound as e:
                    logger.debug(
                        "Loading entry point %s has failed with exception %s", ep, e
                    )
                    continue
            self.__class__.loaded_entry_points = True

        self.transcoder_settings = transcoder_settings
        self.transcoders: Dict[TranscoderName, TranscoderProtocol] = {}
        all_setting_names = self.transcoder_settings.all_names()
        for factory in self.transcoder_factories:
            name = TranscoderName(factory.__name__.lower())
            settings = self.transcoder_settings.for_name(name)
            all_setting_names -= set([name])
            logger.debug(
                "Initializing transcoder %s%s",
                name,
                (
                    " with user-supplied settings %s" % settings
                    if settings
                    else " without any user-supplied settings"
                ),
            )
            transcoder = factory(settings)
            self.transcoders[name] = transcoder
        if all_setting_names:
            raise ValueError(
                "Settings for unavailable transcoders %s have been specified in configuration"
                % ", ".join("%r" % x for x in all_setting_names),
            )

    def map_pipelines(self, src: Path) -> Tuple[nx.MultiDiGraph, List[TranscodingPath]]:
        srctype = FileType.by_name(FileType.from_path(src))
        org_srctype = srctype
        g = nx.MultiDiGraph()

        t2name = dict([(t, tname) for tname, t in self.transcoders.items()])

        types_explored: Dict[FileType, bool] = collections.defaultdict(bool)
        if srctype not in types_explored:
            types_explored[srctype] = False
        while any(not f for f in types_explored.values()):
            srctype = [x for x, y in types_explored.items() if not y][0]
            g.add_node(srctype)
            for tname, t in self.transcoders.items():
                p = Path(os.path.join(src.parent, src.stem + "." + srctype))
                dsttypes = t.can_transcode(p)
                for d in dsttypes:
                    g.add_node(d)
                    g.add_edge(srctype, d, key=t, label=tname)
                    if d not in types_explored:
                        types_explored[d] = False
            types_explored[srctype] = True

        paths = []
        copytranscoder = self.transcoders[TranscoderName("copy")]
        copypath: List[Tuple[FileType, FileType, TranscoderProtocol]] = [
            (
                org_srctype,
                org_srctype,
                copytranscoder,
            )
        ]
        paths.append(copypath)
        for tt in types_explored.keys():
            paths += list(nx.all_simple_edge_paths(g, org_srctype, tt))

        new_paths: List[
            Tuple[int, List[Tuple[FileType, FileType, TranscoderName]]]
        ] = []
        for path in paths:
            length = len(path)
            cost = 0
            new_path = []
            append = True
            for step in path:
                s = step[0]
                dest = step[1]
                transcoder: TranscoderProtocol = step[2]
                if length > 1 and transcoder == copytranscoder:
                    append = False
                    break
                cost += transcoder.cost
                new_path.append((s, dest, t2name[transcoder]))
            if append:
                new_paths.append((cost, new_path))

        final_paths = [
            TranscodingPath(cost, self, path) for cost, path in list(sorted(new_paths))
        ]
        return g, final_paths

    @classmethod
    def register(
        cls, transcoder_factory
    ):  # type: (Type[TranscoderRegistry], Type[TranscoderProtocol]) -> Type[TranscoderProtocol]
        cls.transcoder_factories.add(transcoder_factory)
        return transcoder_factory

    def __str__(self) -> str:
        return "<EncoderRegistry: %s>" % ", ".join(self.transcoders.keys())

    def get_transcoder(self, transcoder_name: TranscoderName) -> TranscoderProtocol:
        return self.transcoders[transcoder_name]


register = TranscoderRegistry.register



import logging
import os
import sys
import typing

from musictoolbox.logging import basicConfig


def main() -> int:
    basicConfig(main_module_name=__name__, level=logging.DEBUG)

    replacements: typing.Dict[str, str] = {}

    for fn in sys.argv[1:]:

        didchange = False

        f = open(fn, "r")
        d = os.path.dirname(fn)
        lines = [x.strip() for x in f.readlines() if x.strip()]

        newlines = []
        for line in lines:
            fullpath = os.path.join(d, line)
            if line.startswith("#"):
                newlines.append(line)
            elif not os.path.exists(fullpath):
                didchange = True
                if fullpath in replacements:
                    newlines.append(replacements[fullpath])
                else:
                    print("%r does not exist" % line)
                    print("Enter new path to replace old one")
                    newline = sys.stdin.readline().strip()
                    if not newline:
                        print("Skipping replacement of %r" % line)
                    newlines.append(newline)
                    replacements[fullpath] = newline
            else:
                newlines.append(line)

        if didchange:
            f = open(fn + ".new", "w")
            f.write("\n".join(newlines))
            f.flush()
            f.close()
            os.rename(fn, fn + "~")
            os.rename(fn + ".new", fn)

    return 0



import argparse
import logging
import os
from pathlib import Path
import subprocess
import sys
import tempfile
from typing import Optional

from networkx import drawing, MultiDiGraph  # type: ignore

from . import config, policies, transcoder, registry
from ..logging import basicConfig

# FIXME type comment
from ..tagging import transfer_tags  # type: ignore
from .interfaces import FileType, TranscoderName


def parser() -> argparse.ArgumentParser:
    p = argparse.ArgumentParser(
        description="Transcodes a media file into another format.",
    )
    p.add_argument("src", help="source file name", type=str)
    p.add_argument("dst", help="destination file name", type=str)
    meg = p.add_mutually_exclusive_group()
    meg.add_argument(
        "-l",
        "--plot",
        help="only plot the available transcoder pipelines",
        action="store_true",
        default=False,
    )
    meg.add_argument(
        "-n",
        "--dry-run",
        help="only display what would be done",
        action="store_true",
        default=False,
    )
    p.add_argument(
        "-p",
        "--pipeline",
        help="comma-separated name of transcoder pipeline steps that overrides the configuration file; possible steps for a (src, dst) pair of files displayed with -n",
        type=str,
        default="",
    )
    p.add_argument(
        "-c",
        "--config-file",
        help="path to a configuration file with transcoding policies and settings; highest precedence config will be loaded from %s; empty string config file will prevent loading any config"
        % config.transcoding_config_default_path(),
        type=str,
        default=None,
    )
    p.add_argument(
        "-d",
        "--debug",
        help="enable debug logging",
        action="store_true",
        default=False,
    )
    return p


def plot_transcoder_pipelines(graph: MultiDiGraph) -> None:
    with tempfile.NamedTemporaryFile() as f:
        drawing.nx_pydot.write_dot(graph, f.name)
        f.seek(0, 0)
        out = subprocess.check_output(["dot", "-Tpng"], stdin=f)
        f.seek(0, 0)
        f.truncate()
        f.write(out)
        f.flush()
        subprocess.check_call(["eog", f.name])


def main() -> Optional[int]:
    p = parser()
    opts = p.parse_args()
    basicConfig(__name__, logging.DEBUG if opts.debug else logging.INFO)

    c = config.load_transcoding_config(opts.config_file)
    r = registry.TranscoderRegistry(c.settings)
    src, dst = Path(opts.src), Path(opts.dst) if opts.dst else None

    pipeline = (
        [TranscoderName(n) for n in opts.pipeline.split(",")] if opts.pipeline else []
    )
    selector = (
        policies.NoPolicyPipelineSelector()
        if pipeline
        else policies.PolicyBasedPipelineSelector(c.policies)
    )
    mapper = transcoder.TranscodingMapper(r, selector)

    dsttype = FileType.by_name(FileType.from_path(dst)) if dst else None
    graph, selected_paths = mapper.lookup_with_graph(
        src, dsttype=dsttype, pipeline=pipeline
    )

    if opts.plot:
        plot_transcoder_pipelines(graph)
        return 0
    elif opts.dry_run:
        for nn, path in enumerate(selected_paths):
            print(
                "Transcoder pipeline %swith cost %s:"
                % ("(selected) " if nn == 0 else "", path.cost)
            )
            for n, step in enumerate(path.steps):
                print("%3d. %s" % (n + 1, step))
            print()
        if not selected_paths:
            print(
                "No transcoder pipelines found for the involved file formats the constraints from configuration or parameters."
            )
        return 4
    elif not dst:
        print(
            "The destination file name cannot be empty if --plot or --dry-run weren't requested.",
            file=sys.stderr,
        )
        return os.EX_USAGE
    elif not selected_paths:
        print(
            "Cannot transcode %s -- no transcoding pipelines found for the involved file formats from configuration or parameters."
            % src,
            file=sys.stderr,
        )
        return 4

    syncer = transcoder.SingleItemSyncer(transfer_tags)
    syncer.sync(src, dst, selected_paths[0])
    return 0



import argparse
import logging
import os

from ..logging import basicConfig


def main() -> int:

    basicConfig(main_module_name=__name__, level=logging.DEBUG)
    p = argparse.ArgumentParser(
        description="Copy an M3U playlist, preserving its relative paths or making them absolute."
    )
    p.add_argument("IN", type=str, nargs="+", help="input playlist")
    p.add_argument("OUT", type=str, nargs=1, help="output playlist")
    p.add_argument(
        "-A",
        "--absolute",
        action="store_true",
        help="make paths in output playlist absolute",
    )

    args = p.parse_args()

    for in_file in args.IN:
        new = []
        in_basedir = os.path.dirname(in_file)
        out_file = args.OUT[0]
        if os.path.isdir(out_file):
            out_basedir = out_file
            out_file = os.path.join(out_basedir, os.path.basename(in_file))
        else:
            assert len(args.IN) == 1, (
                "Error: multiple source paths were specified, but destination %r is not a directory."
                % out_file
            )
            out_basedir = os.path.dirname(out_file)

        with open(in_file, "r") as in_:
            for line in in_:
                if line.startswith("#"):
                    new.append(line)
                    continue
                has_newline = line.endswith("\n")
                line = line.rstrip()
                absline = os.path.abspath(os.path.join(in_basedir, line))
                line = absline
                if not args.absolute:
                    relline = os.path.relpath(line, out_basedir)
                    line = relline
                new.append(line + ("\n" if has_newline else ""))

        try:
            with open(out_file, "r") as out_orig:
                orig_text = out_orig.read()
        except Exception:
            orig_text = ""

        new_text = "".join(new)

        if new_text != orig_text:
            with open(out_file, "w") as out_:
                out_.write(new_text)

    return 0



from pathlib import Path
from typing import List, Dict, Any, Optional


from musictoolbox.transcoding.interfaces import FileType


class BinaryTranscoder(object):
    def can_transcode(self, src: Path) -> List[FileType]:
        name = self.__class__.__name__
        if "Using" in name:
            name = name.split("Using")[0]
        parts = [FileType.by_name(x.lower()) for x in name.split("To")]
        if parts[0] == FileType.from_path(src):
            return [parts[1]]
        return []


class NoSettings(object):
    def __init__(self, settings: Optional[Dict[str, Any]]):
        if settings:
            raise ValueError(
                "the %s transcoder takes no settings" % self.__class__.__name__.lower()
            )


class WithEnvironmentVariables(object):
    def __init__(self, settings: Optional[Dict[str, Any]]):
        if not hasattr(self, "settings"):
            self.settings: Dict[str, Any] = {}
        if settings and "environment_variables" in settings:
            if not "environment_variables" in self.settings:
                self.settings["environment_variables"] = {}
            for k, v in settings["environment_variables"].items():
                if not isinstance(k, str) or not isinstance(v, str):
                    raise ValueError(
                        "The environment variable %r is not a string or its value %r is not a string"
                        % (k, v)
                    )
                self.settings["environment_variables"][k] = v
        else:
            self.settings["environment_variables"] = {}


class WithSettings(object):
    def __init__(self, settings: Optional[Dict[str, Any]]):
        self.settings = settings or {}



from pathlib import Path
from typing import List, Any, Type, Dict, Callable, Union, Protocol  # @UnusedImport


class FileType(str):
    cache = {}  # type: Dict[str, FileType]

    @classmethod
    def by_name(klass, name):  # type: (Type[FileType], str) -> FileType
        if name not in klass.cache:
            klass.cache[name] = FileType(name)
        return klass.cache[name]

    @classmethod
    def from_path(klass, path):  # type: (Type[FileType], Path) -> FileType
        name = path.suffix[1:].lower()
        return klass.by_name(name)


class TranscoderProtocol(Protocol):
    """Interface for transcoders."""

    cost: int = -1

    def __init__(self, settings: Dict[str, Any]):
        pass

    def transcode(self, src: Path, dest: Path) -> None:
        """Transcodes a file src to a file dest."""
        pass

    def can_transcode(self, src: Path) -> List[FileType]:
        pass


class TranscoderName(str):
    pass


class TranscoderLookupProtocol(Protocol):
    def get_transcoder(self, transcoder_name: TranscoderName) -> TranscoderProtocol:
        pass


Postprocessor = Callable[[str, str, Union[None, str], Union[None, str]], None]



from typing import Dict, Any, Set

from musictoolbox.transcoding.interfaces import TranscoderName


class TranscoderSettings(object):
    def __init__(self, settings_dict: Dict[str, Dict[str, Any]]):
        x = (
            dict((x.lower(), y) for x, y in settings_dict.items())
            if settings_dict
            else {}
        )
        self.settings: Dict[str, Any] = x

    def all_names(self) -> Set[TranscoderName]:
        return set(TranscoderName(x) for x in self.settings.keys())

    def for_name(self, name: TranscoderName) -> Dict[str, Any]:
        deff: Dict[str, Any] = {}
        if name in self.settings:
            return self.settings[name]  # type: ignore
        return deff

    def __str__(self) -> str:
        return "<TranscoderSettings settings: %s>" % str(self.settings)



import collections
import os
from pathlib import Path
import tempfile
import typing
import unittest

from . import algo as mod
from ..files import AbsolutePath, Absolutize
from ..transcoding.registry import TranscodingPath
from ..transcoding.test_registry import DummyLookup


def ab(
    iterable: typing.Union[typing.Dict[str, str], typing.List[str], str]
) -> typing.Union[
    typing.Dict[AbsolutePath, AbsolutePath],
    typing.List[AbsolutePath],
    AbsolutePath,
]:
    if isinstance(iterable, list):
        return [Absolutize(x) for x in iterable]
    elif isinstance(iterable, dict):
        return dict(
            (
                Absolutize(x),
                Absolutize(y),
            )
            for x, y in iterable.items()
        )
    elif isinstance(iterable, str):
        return Absolutize(iterable)
    else:
        assert 0, "not reached: %s" % type(iterable)


def abp(x: str) -> AbsolutePath:
    return typing.cast(AbsolutePath, ab(x))


def abd(x: typing.Dict[str, str]) -> typing.Dict[AbsolutePath, AbsolutePath]:
    return typing.cast(typing.Dict[AbsolutePath, AbsolutePath], ab(x))


def abl(x: typing.List[str]) -> typing.List[AbsolutePath]:
    return typing.cast(typing.List[AbsolutePath], ab(x))


class TestFilesystemPathMapper(unittest.TestCase):
    def test_base_case(self) -> None:
        c = mod.FilesystemPathMapper(abp("/mnt/d"))
        c.mptypes = {"/mnt/d": "vfat", "/": "ext4"}

        # Verify that the path is returned exactly the same
        # when it refers to a file in a case-sensitive mount point.
        in_ = abp("/a/b/c/D")
        want = abp("/a/b/c/D")
        got = c.map(in_)
        assert got == want, f"got: {got}, want: {want}"

        # Verify that another path is returned exactly the same
        # when it refers to a file in a case-sensitive mount point.
        in_ = abp("/a/b/c/d")
        want = abp("/a/b/c/d")
        got = c.map(in_)
        assert got == want, f"got: {got}, want: {want}"

    def test_vfat_case(self) -> None:
        c = mod.FilesystemPathMapper(abp("/mnt/d"))
        c.mptypes = {"/mnt/d": "vfat", "/": "ext4"}

        # Verify that the path is returned exactly the same
        # when it refers to a file in a case-sensitive mount point.
        in_ = abp("/mnt/d/x/D")
        want = abp("/mnt/d/x/D")
        got = c.map(in_)
        assert got == want, f"got: {got}, want: {want}"

        # Verify that another path is returned exactly the same
        # when it refers to a file in a case-sensitive mount point.
        in_ = abp("/mnt/d/X/d")
        want = abp("/mnt/d/x/D")
        got = c.map(in_)
        assert got == want, f"got: {got}, want: {want}"

    def test_vfat_illegal(self) -> None:
        c = mod.FilesystemPathMapper(abp("/mnt/d"))
        c.mptypes = {"/mnt/d": "vfat", "/": "ext4"}

        # Verify that the path is returned changed
        # but only for the parts that correspond to the VFAT mountpoint.
        in_ = abp("/mnt/d/Can we sing this song tonight?.mp3")
        want = abp("/mnt/d/Can we sing this song tonight_.mp3")
        got = c.map(in_)
        assert got == want, f"got: {got}, want: {want}"

        # Verify that the path is returned as previously seen
        # but only for the parts that correspond to the VFAT mountpoint.
        in_ = abp("/mnt/d/can we sing this song tonight?.mp3")
        want = abp("/mnt/d/Can we sing this song tonight_.mp3")
        got = c.map(in_)
        assert got == want, f"got: {got}, want: {want}"

        # Verify that parts not corresponding to the mount point
        # are not transmogrified.
        in_ = abp("/mnt/?/Can we sing this song tonight?.mp3")
        want = abp("/mnt/?/Can we sing this song tonight?.mp3")
        got = c.map(in_)
        assert got == want, f"got: {got}, want: {want}"


class TestVfatProtect(unittest.TestCase):
    def test_noendingdots(self) -> None:
        p = mod.vfatprotect("/some/path/with./dots")
        self.assertNotIn(".", p)
        p = mod.vfatprotect("/some/path/with..../many dots")
        self.assertNotIn(".", p)
        p = mod.vfatprotect("/some/path/with /many spaces")
        self.assertNotIn(" /", p)

    def test_nobadchars(self) -> None:
        for f in "?<>\\:*|":
            p = mod.vfatprotect("/path/with%sbadchar" % f)
            self.assertNotIn(f, p)


class AlwaysN(object):
    def __init__(self, ret: int) -> None:
        self.ret = ret

    def compare(self, _: Path, __: Path) -> int:
        return self.ret


DummyTranscodingPath = TranscodingPath(0, DummyLookup(), [])


class _DummyTranscodingPather(object):
    def lookup(self, __arg: Path) -> typing.List[TranscodingPath]:
        return [DummyTranscodingPath]


AlwaysNewer = AlwaysN(1)
AlwaysEqual = AlwaysN(0)
AlwaysOlder = AlwaysN(-1)
DummyTranscodingPather = _DummyTranscodingPather()


class TestComputeSynchronization(unittest.TestCase):
    maxDiff = 65536

    def test_simple_case(self) -> None:
        got = mod.compute_synchronization(
            [],
            abp("/"),
            [],
            abp("/"),
            [],
            [],
            DummyTranscodingPather,
            AlwaysNewer,
        )
        want: mod.SyncRet = ([], {}, {}, [])
        self.assertEqual(got, want)

    def test_identical_and_older(self) -> None:
        got = mod.compute_synchronization(
            abl(["/a"]),
            abp("/"),
            abl(["/target/a"]),
            abp("/target"),
            [],
            [],
            DummyTranscodingPather,
            AlwaysEqual,
        )
        want: mod.SyncRet = ([], {}, abd({"/a": "/target/a"}), [])
        self.assertEqual(got, want)

        got = mod.compute_synchronization(
            abl(["/a"]),
            abp("/"),
            abl(["/target/a"]),
            abp("/target"),
            [],
            [],
            DummyTranscodingPather,
            AlwaysOlder,
        )
        self.assertEqual(got, want)

    def test_source_is_newer_than_target(self) -> None:
        got = mod.compute_synchronization(
            abl(["/a"]),
            abp("/"),
            abl(["/target/a"]),
            abp("/target"),
            [],
            [],
            DummyTranscodingPather,
            AlwaysNewer,
        )
        want: mod.SyncRet = (
            [
                (abp("/a"), abp("/target/a"), DummyTranscodingPath),
            ],
            {},
            {},
            [],
        )
        self.assertEqual(got, want)

    def test_absent_target(self) -> None:
        got = mod.compute_synchronization(
            abl(["/a"]),
            abp("/"),
            [],
            abp("/target"),
            [],
            [],
            DummyTranscodingPather,
            AlwaysNewer,
        )
        want: mod.SyncRet = (
            [
                (abp("/a"), abp("/target/a"), DummyTranscodingPath),
            ],
            {},
            {},
            [],
        )
        self.assertEqual(got, want)

    def test_absent_source(self) -> None:
        got = mod.compute_synchronization(
            [],
            abp("/"),
            abl(["/target/a"]),
            abp("/target"),
            [],
            [],
            DummyTranscodingPather,
            AlwaysEqual,
        )
        want: mod.SyncRet = ([], {}, {}, abl(["/target/a"]))
        self.assertEqual(got, want)

    def test_mapper(self) -> None:
        c = mod.FilesystemPathMapper(abp("/mnt/d"))
        c.mptypes = {"/mnt/d": "vfat", "/": "ext4"}

        got = mod.compute_synchronization(
            abl(["/basedir/a"]),
            abp("/basedir"),
            abl(["/mnt/d/target/a"]),
            abp("/mnt/d/target"),
            [],
            [c],
            DummyTranscodingPather,
            AlwaysNewer,
        )
        want: mod.SyncRet = (
            [
                (
                    abp("/basedir/a"),
                    abp("/mnt/d/target/a"),
                    DummyTranscodingPath,
                ),
            ],
            {},
            {},
            [],
        )
        self.assertEqual(got, want)

    def test_mapper_conflicts(self) -> None:
        c = mod.FilesystemPathMapper(abp("/mnt/d"))
        c.mptypes = {"/mnt/d": "vfat", "/": "ext4"}

        got = mod.compute_synchronization(
            abl(["/basedir/a", "/basedir/A"]),
            abp("/basedir"),
            abl(["/mnt/d/target/a"]),
            abp("/mnt/d/target"),
            [],
            [c],
            DummyTranscodingPather,
            AlwaysNewer,
        )
        want: mod.SyncRet = (
            [
                (
                    abp("/basedir/a"),
                    abp("/mnt/d/target/a"),
                    DummyTranscodingPath,
                ),
            ],
            collections.OrderedDict(
                {
                    abp("/basedir/A"): mod.Conflict(
                        abp("/basedir/A"),
                        abp("/mnt/d/target/a"),
                        abp("/basedir/a"),
                    )
                }
            ),
            {},
            [],
        )
        assert got[1] == want[1]

    def test_complex(self) -> None:
        c = mod.FilesystemPathMapper(abp("/mnt/d"))
        c.mptypes = {"/mnt/d": "vfat", "/": "ext4"}

        class C(object):
            def compare(self, p1: Path, unused_p2: Path) -> int:
                if p1.name == "newerintgt":
                    return -1
                return 1

        got = mod.compute_synchronization(
            abl(
                [
                    "/basedir/needsupdate",
                    "/basedir/differentname",
                    "/basedir/newerintgt",
                    "/basedir/absentintgt",
                ]
            ),
            abp("/basedir"),
            abl(
                [
                    "/mnt/d/target/needsupdate",
                    "/mnt/d/target/DifferentName",
                    "/mnt/d/target/absentinsource",
                ]
            ),
            abp("/mnt/d/target"),
            [],
            [c],
            DummyTranscodingPather,
            C(),
        )
        want: mod.SyncRet = (
            [
                (
                    abp("/basedir/needsupdate"),
                    abp("/mnt/d/target/needsupdate"),
                    DummyTranscodingPath,
                ),
                (
                    abp("/basedir/differentname"),
                    abp("/mnt/d/target/DifferentName"),
                    DummyTranscodingPath,
                ),
                (
                    abp("/basedir/absentintgt"),
                    abp("/mnt/d/target/absentintgt"),
                    DummyTranscodingPath,
                ),
            ],
            {},
            abd(
                {
                    "/basedir/newerintgt": "/mnt/d/target/newerintgt",
                }
            ),
            [abp("/mnt/d/target/absentinsource")],
        )
        self.assertEqual(got, want)


class TestModTimestampComparer(unittest.TestCase):
    def test_regular(self) -> None:
        c = mod.ModTimestampComparer()
        with tempfile.TemporaryDirectory() as d:
            f1 = Absolutize(d) / "f1"
            f2 = Absolutize(d) / "f2"
            with open(f1, "w") as f1o:
                f1o.write("x")
            with open(f2, "w") as f2o:
                f2o.write("y")

            os.utime(f1, (f2.stat().st_mtime + 1, f2.stat().st_mtime + 1))
            assert c.compare(f1, f2) == 1

            os.utime(f1, (f2.stat().st_mtime - 1, f2.stat().st_mtime - 1))
            assert c.compare(f1, f2) == -1

            os.utime(f1, (f2.stat().st_mtime, f2.stat().st_mtime))
            assert c.compare(f1, f2) == 0

    def test_vfat(self) -> None:
        c = mod.ModTimestampComparer()
        with tempfile.TemporaryDirectory() as d:
            c.mptypes[d] = "vfat"
            f1 = Absolutize(d) / "f1"
            f2 = Absolutize(d) / "f2"
            with open(f1, "w") as f1o:
                f1o.write("x")
            with open(f2, "w") as f2o:
                f2o.write("y")

            os.utime(f1, (f2.stat().st_mtime + 1, f2.stat().st_mtime + 1))
            assert c.compare(f1, f2) == 0

            os.utime(f1, (f2.stat().st_mtime - 1, f2.stat().st_mtime - 1))
            assert c.compare(f1, f2) == 0

            os.utime(f1, (f2.stat().st_mtime, f2.stat().st_mtime))
            assert c.compare(f1, f2) == 0


class TestWithin(unittest.TestCase):
    def test_same_matches(self) -> None:
        assert mod.within(Absolutize("/a"), Absolutize("/a/"))

    def test_subpath_matches(self) -> None:
        assert mod.within(Absolutize("/a"), Absolutize("/a/bcd"))

    def test_non_subpaths_do_not_match(self) -> None:
        assert not mod.within(Absolutize("/a"), Absolutize("/bcd"))
        assert not mod.within(Absolutize("/a"), Absolutize("/a/../bcd"))



#!/usr/bin/python3

import glob
import os

from setuptools import setup


directory = os.path.dirname(__file__)
path_to_main_file = os.path.join(directory, "lib/musictoolbox/__init__.py")
path_to_readme = os.path.join(directory, "README.md")
for line in open(path_to_main_file):
    if line.startswith("__version__"):
        version = line.split()[-1].strip("'").strip('"')
        break
else:
    raise ValueError('"__version__" not found in "lib/musictoolbox/__init__.py"')
readme = open(path_to_readme).read(-1)

classifiers = [
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Intended Audience :: End Users/Desktop",
    "License :: OSI Approved :: GNU General Public License (GPL)",
    "Operating System :: POSIX :: Linux",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.7",
    "Topic :: Utilities",
]

setup(
    name="musictoolbox",
    version=version,
    description="A set of utilities to help you groom your music collection",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Manuel Amador (Rudd-O)",
    author_email="rudd-o@rudd-o.com",
    license="GPL",
    url="http://github.com/Rudd-O/musictoolbox",
    package_dir=dict(
        [
            ("musictoolbox", "lib/musictoolbox"),
        ]
    ),
    classifiers=classifiers,
    packages=["musictoolbox",
              "musictoolbox.sync",
              "musictoolbox.transcoding",
              "musictoolbox.transcoding.codecs",
              "musictoolbox.cmd"],
    install_requires=[
        "mutagen",
        "Twisted",
        "packaging",
        "networkx",
        "pyxdg",
        "psutil",
    ],
    entry_points={
        "musictoolbox.transcoding.codecs": [
            "copy = musictoolbox.transcoding.codecs.basic:Copy",
        ]
    },
    scripts=[f for f in glob.glob(os.path.join("bin", "*")) if not f.endswith("~")],
    keywords="mp3 ogg mkv transcoding aac mp4 video flv flac",
    zip_safe=False,
)



from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
import logging
import os
import sys
import textwrap
import typing

from ..tagging import transfer_tags  # type: ignore

from ..files import Absolutize
from ..logging import basicConfig
from ..transcoding import config, policies, registry
from ..transcoding.interfaces import Postprocessor
from ..transcoding.transcoder import TranscodingMapper
from .core import Synchronizer


logger = logging.getLogger(__name__)


class SynchronizationCLI:

    debug = False

    def __init__(
        self,
        destpath: str,
        playlists: typing.List[str],
        dryrun: bool,
        concurrency: int,
        delete: bool,
        exclude_beneath: typing.List[str],
        transcoding_mapper: TranscodingMapper,
        postprocessor: Postprocessor,
        force_vfat: bool,
    ) -> None:
        self.synchronizer = Synchronizer(
            [Absolutize(p) for p in playlists],
            Absolutize(destpath),
            transcoding_mapper,
            [Absolutize(p) for p in exclude_beneath],
            postprocessor,
            force_vfat,
        )
        self.dryrun = dryrun
        self.delete = delete
        self.concurrency = concurrency

    def run(self) -> int:
        """Returns:

        0 if all transcoding / synchronization operations completed.
        2 if scanning experienced a problem.
        4 if a transcoding / synchronization failure took place.
        8 if a problem took place while writing playlists.
        16 if a problem took place while removing files.
        Or a combination of the above.

        Raises Exception in an unexpected failure took place.

        If self.debug == True (see __init__), raises the exceptions
        that the transcoding / sync operations saw.
        """
        try:
            sync_plan = self.synchronizer.compute_synchronization()
        except Exception:
            logger.exception("Error scanning source material")
            return 2

        will_sync, cant_sync, already_synced, will_delete = sync_plan

        for src, dst in already_synced.items():
            logger.info("No need to sync %s — already synced to %s", src, dst)

        for src, dst, _ in will_sync:
            logger.info("Will sync %s to %s", src, dst)

        for src, ex in cant_sync.items():
            logger.info("Cannot sync %s: %s", src, ex)

        if self.delete:
            for fn in will_delete:
                logger.info("Will delete %s", fn)

        problems_syncing = False

        if not self.dryrun:
            q, cancel = self.synchronizer.synchronize(sync_plan, self.concurrency)
            try:
                while True:
                    r = q.get(block=True)
                    if r is None:
                        break
                    s, d, exc = r
                    if exc:
                        problems_syncing = True
                        logger.error("Could not sync file %s: %s", s, exc)
                    else:
                        logger.info("Synced file: %s -> %s", s, d)
            except KeyboardInterrupt:
                cancel()
                raise

        problems_playlisting = False

        for oldp, newp, exc in self.synchronizer.synchronize_playlists(
            sync_plan, self.dryrun
        ):
            if exc:
                problems_playlisting = True
                logger.error("Could not sync playlist %s: %s", newp, exc)
            else:
                if not self.dryrun:
                    logger.info("Synced playlist: %s -> %s", oldp, newp)

        problems_deleting = False

        if self.delete:
            for p, exc in self.synchronizer.synchronize_deletions(
                sync_plan, self.dryrun
            ):
                if exc:
                    problems_deleting = True
                    logger.error("Could not remove file %s: %s", p, exc)
                else:
                    if not self.dryrun:
                        logger.info("Removed file: %s", p)

        retval = 0
        if problems_syncing or cant_sync:
            retval += 4
        if problems_playlisting:
            retval += 8
        if problems_deleting:
            retval += 16
        return retval


def get_parser() -> ArgumentParser:
    program_name = os.path.basename(sys.argv[0])
    program_shortdesc = "Synchronizes music to a directory."
    program_shortdesc += (
        "\n\n"
        + textwrap.dedent(
            """
This program takes a number of file lists / playlists in the command line,
and a destination directory, then synchronizes all the songs in the playlists
to the destination directory, with optional modifications to the files and their
names as they are copied to the destination directory, for example preserving
the path structure except for changing characters incompatible with VFAT
volumes to underscores.

For example, suppose you have two playlists `a.m3u` and `b.m3u` (perhaps
generated with the companion `genplaylist` program) that contain the
respective following song lists:

1. /shared/Music/Ace of Base/Everytime it rains.mp3
2. /shared/Music/Scatman John/Scatman.mp3

1. /shared/Music/Dr. Alban/It's my life (12" mix).mp3

If you run `syncplaylists a.m3u b.m3u "/mounteddrive/Travel Music"`, the
resulting directory structure in `/mounteddrive/Travel Music` will look
like this:

1. /mounteddrive/Travel Music/Ace of Base/Everytime it rains.mp3
2. /mounteddrive/Travel Music/Scatman John/Scatman.mp3
3. /mounteddrive/Travel Music/Music/Dr. Alban/It's my life (12_ mix).mp3
4. /mounteddrive/Travel Music/Playlists/a.m3u
5. /mounteddrive/Travel Music/Playlists/b.m3u

with the paths in the brand new playlists `a.m3u` and `b.m3u` transmuted
so that the song paths are relative to the location of the playlists.  This way
your media player can automatically use the playlists you specified to play
music for you.

`syncplaylists` also has the ability to transcode the files to the preferred
formats of your target device as well.  `syncplaylists will read an INI file
named `~/.syncplaylists.ini` to determine how to transcode.  For example,
if you wanted to transcode all non-MP3 files to MP3, you could add the
following configuration to the file:

    [transcoding]
    mp3=copy
    *=mp3

and then `syncplaylists` will do a best-effort attempt to transcode the files
in your playlists to MP3, except for the MP3 files themselves which would just
get copied as per the `copy` action specified in the configuration file.
    """
        ).strip()
    )

    parser = ArgumentParser(
        prog=program_name,
        description=program_shortdesc,
        formatter_class=RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-n",
        "--dry-run",
        dest="dryrun",
        action="store_true",
        help="do nothing except show what will happen [default: %(default)s]",
    )
    parser.add_argument(
        "-d",
        "--delete",
        dest="delete",
        action="store_true",
        help="remove all files that are not mentioned in the playlists [default: %(default)s]",
    )
    parser.add_argument(
        "-D",
        "--debug",
        dest="debug",
        action="store_true",
        help="raise exceptions instead of reporting them in a succinct manner [default: %(default)s]",
    )
    parser.add_argument(
        "-e",
        "--exclude",
        dest="exclude",
        action="append",
        default=[],
        help="ignore all files underneath these folders [default: %(default)s]",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="count",
        help="set verbosity level [default: %(default)s]",
    )
    parser.add_argument(
        "-V",
        "--force-vfat",
        dest="force_vfat",
        action="store_true",
        default=False,
        help="assume that the destination folder is stored on a FAT file system, and perform the path name conversions appropriate for the case -- useful to sync files to Android devices or other typical music players [default is to autodetect based on the destination mount point]",
    )
    parser.add_argument(
        "--concurrency",
        metavar="NUMPROCS",
        dest="concurrency",
        type=int,
        default=-1,
        help="number of concurrent processes to run [default: automatic]",
    )
    parser.add_argument(
        dest="playlists",
        help="paths to M3U playlists to synchronize",
        metavar="playlist",
        nargs="+",
    )
    parser.add_argument(
        "-c",
        "--config-file",
        help="path to a configuration file with transcoding policies and settings; highest precedence config will be loaded from %s"
        % config.transcoding_config_default_path(),
        type=str,
        default=None,
    )
    parser.add_argument(
        "-p",
        "--profile-file",
        help="path to a file where cProfile output will be stored",
        type=str,
        default=None,
    )
    parser.add_argument(dest="destpath", help="destination directory", metavar="dir")
    return parser


def run_sync(
    dryrun: bool,
    destpath: str,
    playlists: typing.List[str],
    concurrency: int,
    delete: bool,
    exclude_beneath: typing.List[str],
    configfile: typing.Optional[str] = None,
    profilefile: typing.Optional[str] = None,
    force_vfat: typing.Optional[bool] = False,
) -> int:
    """Runs sync process.  Returns what SynchronizationCLIBackend.run() does."""
    cfg = config.load_transcoding_config(configfile)
    reg = registry.TranscoderRegistry(cfg.settings)
    sel = policies.PolicyBasedPipelineSelector(cfg.policies, allow_fallback=False)
    tm = TranscodingMapper(reg, sel)
    pp = transfer_tags

    def w() -> int:
        return SynchronizationCLI(
            destpath,
            playlists,
            dryrun,
            concurrency,
            delete,
            exclude_beneath,
            tm,
            pp,
            force_vfat or False,
        ).run()

    if profilefile:
        import cProfile

        cProfile.runctx("w()", {}, locals(), filename=profilefile)
        return 0

    ret: int = w()
    return ret


def main(argv: typing.Optional[typing.List[str]] = None) -> int:
    """Command line options."""

    if argv is None:
        argv = sys.argv
    else:
        sys.argv.extend(argv)

    parser = get_parser()
    args = parser.parse_args()

    if args.debug:
        level = logging.DEBUG
    elif args.verbose:
        level = logging.INFO
    else:
        level = logging.WARNING
    basicConfig(__name__, level)

    return run_sync(
        dryrun=args.dryrun,
        playlists=args.playlists or [],
        concurrency=args.concurrency,
        destpath=args.destpath,
        delete=args.delete,
        exclude_beneath=args.exclude or [],
        configfile=args.config_file,
        profilefile=args.profile_file,
        force_vfat=args.force_vfat,
    )


if __name__ == "__main__":
    sys.exit(main())



import typing

from ..files import AbsolutePath


class PathMappingProtocol(typing.Protocol):
    def map(self, __arg: AbsolutePath) -> AbsolutePath:
        pass


class PathComparisonProtocol(typing.Protocol):
    def compare(self, __arg1: AbsolutePath, __arg2: AbsolutePath) -> int:
        pass



import collections
import fcntl
import io
import logging
import collections.abc
import os
import pickle
import typing
import xdg.BaseDirectory


_LOGGER = logging.getLogger(__name__)


class OnDiskCacheable(typing.Protocol):
    """
    The protocol for on-disk cacheable objects.

    Objects implementing this protocol must be pickleable.
    """

    def is_dirty(self) -> bool:
        """
        Is the cache dirty?

        Implementors must keep track of such state and return accordingly.
        """
        pass

    def mark_clean(self) -> None:
        """
        Mark the cache clean.

        Implementors must keep track of such state.
        """
        pass


C = typing.TypeVar("C")


class FileMetadataCache(OnDiskCacheable, typing.Generic[C]):
    def __init__(self, cache_item_factory: collections.abc.Callable[[str], C]) -> None:
        """
        Initializes a cache.

        The cache_item_factory takes a path (in string form) and must return a cache
        item, or None if the factory cannot produce an item.

        Cache keys are absolute paths internally.  This is an implementation detail,
        and it is subject to change in the future.  For convenience, the cache store
        is exposed as self._store, but your code will break if you use this directly,
        so limit yourself to the methods exposed by this class.
        """
        self.__factory = cache_item_factory
        self._store: dict[str, tuple[float, C]] = {}
        self.__dirty = False

    def update_metadata_for(self, allfiles: list[str]) -> None:
        """
        Instruct the cache to update itself for the passed files.

        is_dirty() will return True after this, if any of the files passed
        had its corresponding cache entry updated.
        """
        dirty = False
        for ff in allfiles:
            ff = os.path.abspath(ff)
            try:
                modtime = os.stat(ff).st_mtime
            except Exception as exc:
                _LOGGER.error("Error examining %s: %s", ff, exc)
                continue
            if ff in self._store and self._store[ff][0] >= modtime:
                if _LOGGER.level <= logging.DEBUG:
                    _LOGGER.debug("No need to update cache for file %s", ff)
                continue
            if _LOGGER.level <= logging.DEBUG:
                _LOGGER.debug("Updated cache for file %s at mod time %s", ff, modtime)
            metadata = self.__factory(ff)
            self._store[ff] = (modtime, metadata)
            dirty = True
        self.__dirty = dirty

    def __getitem__(self, key: str) -> C:
        key = os.path.abspath(key)
        return self._store[key][1]

    def has_key(self, key: str) -> bool:
        """Returns whether the cache contains an entry for the key."""
        key = os.path.abspath(key)
        return key in self._store

    def get(self, key: str) -> C | None:
        """Return a cache entry."""
        key = os.path.abspath(key)
        m = self._store.get(key)
        if m is None:
            return None
        return m[1]

    def mark_clean(self) -> None:
        """Mark the cache as clean again."""
        self.__dirty = False

    def is_dirty(self) -> bool:
        """Return whether the cache is dirty."""
        return self.__dirty


D = typing.TypeVar("D", bound="OnDiskCacheable")


class OnDiskMetadataCache(typing.Generic[D]):
    """
    Cache utility with locking.

    The object manufactured by the cache_factory is returned when called
    as a context manager, but only after the cache file has safely locked
    on disk.

    The the cache is pickled to the file, and the file is unlocked and
    closed, when the scope of the context manager ends.

    Failing to lead the cache is never an error (it defaults to the
    empty cache object created by the factory), but failing to save the
    cache will raise the appropriate exception.
    """

    def __init__(
        self,
        cache_name: str,
        cache_version: int,
        cache_factory: collections.abc.Callable[[], D],
    ):
        """
        Context manager that initializes an on-disk cache.

        Caches are loaded from $XDG_DATA_DIR/.cache/musictoolbox and named
        as files therein..

        The cache_factory will be called to produce an empty cache in case
        the cache cannot be loaded from disk.
        """
        self.__cache_version = cache_version
        self.__cache_factory = cache_factory
        self.__f: io.BufferedReader | None = None
        self.__metadata: D = None  # type: ignore
        p = xdg.BaseDirectory.save_cache_path("musictoolbox")
        self.__path = os.path.join(p, cache_name.replace(os.path.sep, "_"))

    def __enter__(self) -> D:
        f: io.BufferedReader | None = None

        metadata = self.__cache_factory()
        fsize = 0
        try:
            if _LOGGER.level <= logging.DEBUG:
                _LOGGER.debug("Loading cache from %s", self.__path)
            f = open(self.__path, "a+b")
            fcntl.flock(f, fcntl.LOCK_EX)
            fsize = os.stat(self.__path).st_size
            f.seek(0, 0)
        except Exception as exc:
            if not isinstance(exc, FileNotFoundError):
                _LOGGER.error("Error opening cache from %s: %s", self.__path, exc)

        if f and fsize:
            try:
                version, loaded_metadata = typing.cast(tuple[int, D], pickle.load(f))
                if version == self.__cache_version:
                    metadata = loaded_metadata
                else:
                    _LOGGER.debug(
                        "Expected cache version was %s, loaded cache version was %s,"
                        " ignoring cache",
                        self.__cache_version,
                        version,
                    )
            except Exception as exc:
                _LOGGER.error("Error opening cache from %s: %s", self.__path, exc)

        self.__f = f
        self.__metadata = metadata
        return metadata

    def __exit__(self, *unused_args: typing.Any, **unused_kw: typing.Any) -> None:

        if self.__f and self.__metadata.is_dirty():
            try:
                self.__metadata.mark_clean()
                if _LOGGER.level <= logging.DEBUG:
                    _LOGGER.debug("Saving cache to %s", self.__path)
                self.__f.seek(0, 0)
                self.__f.truncate()
                pickle.dump((self.__cache_version, self.__metadata), self.__f)
                self.__f.flush()
            finally:
                self.__f.close()
